"""
Created on Sat Mar 26 14:24:27 2022

@author: Haotian Teng
"""

import h5py
from xron.utils import seq_op
import numpy as np
from umap import UMAP
from sklearn.manifold import TSNE
from matplotlib import pyplot as plt
import matplotlib.cm as cm
from typing import List
        
fast5_f = "/home/heavens/Documents/FAQ15457_pass_6cc85380_1.fast5"
root = h5py.File(fast5_f,mode = 'r')
k = 5
strides = 11
for read_id in root:
    read_h = root[read_id]
    basecalled = read_h['Analyses/Basecall_1D_001/BaseCalled_template']
    hidden = np.asarray(basecalled['Hidden'])
    signal = np.asarray(read_h['Raw/Signal'])
    move = basecalled['Move']
    position_sig = np.repeat(np.cumsum(move)-1,repeats = strides).astype(np.int32)[:len(signal)]
    segmentation = np.where(np.diff(position_sig))[0]
    position = np.cumsum(move)
    fastq = np.asarray(basecalled['Fastq']).item().decode().split()[1]
    kmers_list = seq_op.kmers2array(seq_op.seq2kmers(fastq,k = k))
    # kmers_list = [0]*((k-1)//2) + kmers_list + [0]*((k-1)//2)
    kmers_list = [0]*(k-1) + kmers_list
    logit = 1/(1+1/np.exp(hidden))
    
    ## Mask to display partail
    max_pos = 10
    hidden = hidden[position<max_pos,:]
    logit = logit[position<max_pos,:]
    position = position[position<max_pos]
    
    size = 4
    umap = UMAP(n_components=2)
    tsne = TSNE(n_components=2, init = 'random')
    hidden_umap = umap.fit_transform(hidden)
    hidden_tsne = tsne.fit_transform(hidden)
    logit_umap = umap.fit_transform(logit)
    logit_tsne = tsne.fit_transform(logit)
    # cla = [kmers_list[x-1] for x in position]
    cla = position
    
    plt.figure(figsize = (80,40))
    max_T = 2000
    plt.plot(signal[::-1][:max_T],color = "red")
    plt.vlines(x = segmentation[segmentation<max_T], ymin = min(signal), ymax = max(signal), color = "green", linestyles = "dotted", linewidth = 0.01, antialiased=False)
    plt.figure()
    scatter = plt.scatter(hidden_umap[:,0],hidden_umap[:,1],s = size,c = cla, cmap='viridis')
    plt.title("hidden umap")
    plt.legend(handles=scatter.legend_elements()[0], labels=list(np.arange(max(cla+1))))
    
    plt.figure()
    scatter = plt.scatter(hidden_tsne[:,0],hidden_tsne[:,1],s = size,c = cla, cmap='viridis')
    plt.title("hidden TSNE")
    plt.legend(handles=scatter.legend_elements()[0], labels=list(np.arange(max(cla+1))))
    
    plt.figure()
    scatter = plt.scatter(logit_umap[:,0],logit_umap[:,1],s = size,c = cla, cmap='viridis')
    plt.title("logit umap")
    plt.legend(handles=scatter.legend_elements()[0], labels=list(np.arange(max(cla+1))))
    
    plt.figure()
    scatter = plt.scatter(logit_tsne[:,0],logit_tsne[:,1],s = size,c = cla, cmap='viridis')
    plt.title("logit TSNE")
    plt.legend(handles=scatter.legend_elements()[0], labels=list(np.arange(max(cla+1))))
    break


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul  1 07:32:52 2022

@author: haotian teng
"""

import os
import sys
import toml
import numpy as np
import argparse

def split(args):
    collections = {}
    fs = os.listdir(args.input)
    for f in fs:
        if f.endswith(".npy"):
            try:
                collections[f] = np.load(os.path.join(args.input,f))
            except ValueError:
                continue
    s = np.array([len(x) for x in collections.values()])
    if not(np.all(s==s[0])):
        print("Warning the npy files inside the folder has different size.")
    for i in np.arange(0,max(s),args.batch_size):
        sub_f = os.path.join(args.input,args.prefix+str(i//args.batch_size))
        os.makedirs(sub_f,exist_ok=True)
        for f,data in collections.items():
            np.save(os.path.join(sub_f,f),data[i:i+args.batch_size])

def merge(args):
    fs = os.listdir(args.input)
    fs = [x for x in fs if args.prefix in x]
    i = 0
    collections = {}
    while args.prefix+str(i) in fs:
        sub_f = os.path.join(args.input,args.prefix + str(i))
        for f in os.listdir(sub_f):
            if f.endswith(".npy"):
                pieces = np.load(os.path.join(sub_f,f))
                if f in collections.keys():
                    collections[f].append(pieces)
                else:
                    collections[f] = [pieces]
        i+=1
    for key in collections.keys():
        print("merge %s"%(key))
        if collections[key][0].ndim > 1:
            np.save(os.path.join(args.input,key),np.vstack(tuple(collections[key])))
        else:
            np.save(os.path.join(args.input,key),np.hstack(tuple(collections[key])))

def main(args):
    if args.merge:
        merge(args)
    else:
        split(args)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training RHMM model')
    parser.add_argument("-i","--input", type = str, required = True,
                        help = "Data folder contains the chunk, kmer sequence.")
    parser.add_argument("-b","--batch_size", type = int, default = 4000,
                        help = "The batch size of each subfolder.")
    parser.add_argument("--prefix", type = str, default = "subdata",
                        help = "The prefix of the sub-folders.")
    parser.add_argument("--reverse",action = "store_true", dest = "merge",
                        help = "Reverse the split operation, surrogate the data.")
    args = parser.parse_args(sys.argv[1:])
    main(args)



"""
Created on Sun Dec 26 19:38:25 2021

@author: Haotian Teng
"""
import os
import sys
import torch
import argparse
from functools import partial
from xron.xron_model import CONFIG,DECODER_CONFIG,CRITIC_CONFIG,MM_CONFIG
from xron.xron_train_supervised import main as supervised_train
from xron.xron_train_variational import main as reinforce_train
from xron.xron_train_embedding import main as embedding_train

optimizers = {'Adam':torch.optim.Adam,
              'AdamW':torch.optim.AdamW,
              'SGD':torch.optim.SGD,
              'RMSprop':torch.optim.RMSprop,
              'Adagrad':torch.optim.Adagrad,
              'Momentum':partial(torch.optim.SGD,momentum = 0.9)}

def main(args):
    class CTC_CONFIG(MM_CONFIG):
        CTC = {"beam_size":5,
               "beam_cut_threshold":0.05,
               "alphabeta": "ACGTM",
               "mode":"rna"}
        
    class TRAIN_EMBEDDING_CONFIG(CTC_CONFIG):
        TRAIN = {"inital_learning_rate":args.lr,
                 "batch_size":args.batch_size,
                 "grad_norm":2,
                 "epsilon":0.1,
                 "epsilon_decay":0,
                 "alpha":0.01, #Entropy loss scale factor
                 "keep_record":5,
                 "decay":args.decay,
                 "diff_signal":args.diff}
    
    class TRAIN_SUPERVISED_CONFIG(CTC_CONFIG):
        TRAIN = {"inital_learning_rate":args.lr,
                 "batch_size":args.batch_size,
                 "grad_norm":2,
                 "keep_record":5,
                 "eval_size":10000,
                 "optimizer":optimizers[args.optimizer]}
    
    class TRAIN_REINFORCE_CONFIG(CTC_CONFIG):
        TRAIN = {"inital_learning_rate":args.lr,
                 "batch_size":args.batch_size,
                 "grad_norm":2,
                 "epsilon":0.1,
                 "epsilon_decay":0,
                 "alpha":0.01, #Entropy loss scale factor
                 "beta": 1., #Reconstruction loss scale factor
                 "gamma":0, #Alignment loss scale factor
                 "preheat":5000,
                 "keep_record":5,
                 "decay":args.decay,
                 "diff_signal":args.diff}
    train_config = {"Embedding":TRAIN_EMBEDDING_CONFIG,
                    "Supervised":TRAIN_SUPERVISED_CONFIG,
                    "Reinforce":TRAIN_REINFORCE_CONFIG}
    train_module = {"Embedding":embedding_train,
                    "Supervised":supervised_train,
                    "Reinforce":reinforce_train}
    args.config = train_config[args.module]
    train_module[args.module](args)

def add_arguments(parser):
    parser.add_argument('--module', required = True,
                        help = "The training module to call, can be Embedding, Supervised and Reinforce")
    parser.add_argument('-i', '--chunks', required = True,
                        help = "The .npy file contain chunks.")
    parser.add_argument('-o', '--model_folder', required = True,
                        help = "The folder to save folder at.")
    parser.add_argument('--seq', required = True,
                        help="The .npy file contain the sequence.")
    parser.add_argument('--seq_len', required = True,
                        help="The .npy file contain the sueqnece length.")
    parser.add_argument('--device', default = 'cuda',
                        help="The device used for training, can be cpu or cuda.")
    parser.add_argument('--lr', default = 4e-3, type = float,
                        help="Initial learning rate.")
    parser.add_argument('--batch_size', default = 200, type = int,
                        help="Training batch size.")
    parser.add_argument('--epoches', default = 10, type = int,
                        help = "The number of epoches to train.")
    parser.add_argument('--report', default = 20, type = int,
                        help = "The interval of training rounds to report.")
    parser.add_argument('--load', dest='retrain', action='store_true',
                        help='Load existed model.')
    parser.add_argument('--config', default = None,
                        help = "Training configuration.")
    parser.add_argument('--optimizer', default = "RMSprop",
                        help = "Optimizer to use, can be Adam, AdamW, SGD and RMSprop,\
                            default is RMSprop")
    parser.add_argument('--threads', type = int, default = None,
                        help = "Number of threads used by Pytorch")

def post_args(args):
    os.makedirs(args.model_folder,exist_ok=True)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Calling training module.')
    add_arguments(parser)
    args = parser.parse_args(sys.argv[1:])
    main(args)


"""
Created on Wed Apr 13 02:41:15 2022

@author: Haotian Teng
"""
import os
import sys
import toml
import argparse
import itertools
import numpy as np
from tqdm import tqdm
from typing import List, Union
from xron.utils.seq_op import fast5_iter,med_normalization,dwell_normalization,combine_normalization,norm_by_noisiest_section

class Extractor(object):
    def __init__(self,k,alphabeta):
        self.k = k
        self.ab_dict = {x:idx for idx,x in enumerate(alphabeta)}
        self.ba_dict = {idx:x for idx,x in enumerate(alphabeta)}
        self.N_BASE = len(alphabeta)
        self.kmer2idx_dict,self.idx2kmer = self.build_kmer_dict()
        
    def kmer_decode(self,
                    sequence:str,
                    signal:np.array,
                    seq_pos:np.array,
                    padded:bool = False) -> List[str]:
        """
        Generate a kmer sequence given the sequence and the relavant sequence
        position.

        Parameters
        ----------
        sequence : str
            The sequence.
        signal : np.array
            The corresponding reversed signal related to the ref_sig_idx.
        seq_pos : np.array
            A array with same length of the sequence, ith element P_i means the
            ith base in sequence is at P_i location in the signal.
        padded : bool, default is False, if we want to pad the kmer sequence.

        Returns
        -------
        kmer_seq : List[str]
            Return the list of the kmers.

        """
        kmer_seq = []
        curr_kmer = ''
        kmer_range = seq_pos[(self.k-1)//2:-(self.k//2-1)]
        seq_duration = seq_pos[1:] - seq_pos[:-1]
        seq_duration = seq_duration[(self.k-1)//2:len(sequence) - self.k//2]
        seq_duration = self._smooth_zero(seq_duration)
        segmented_signal = signal[kmer_range[0]:kmer_range[-1]] if not padded else signal
        for idx in np.arange((self.k-1)//2,len(sequence) - self.k//2):
            start = idx-(self.k-1)//2
            curr_kmer = sequence[start:start+self.k]
            kmer_seq += [self.kmer2idx_dict[curr_kmer]]*seq_duration[idx-(self.k-1)//2]
        kmer_seq = np.asarray(kmer_seq)
        if padded:
            kmer_seq = np.pad(kmer_seq,(kmer_range[0],len(segmented_signal)-kmer_range[-1]),'edge')
        assert(len(segmented_signal) == len(kmer_seq))
        return segmented_signal,kmer_seq
    
    def _smooth_zero(self,duration_vector:np.array):
        """
        Smooth out the zero duration to prevent >1 jump in kmer train.

        Parameters
        ----------
        duration_vector : np.array
            The duration vector.

        """
        zero_count = (duration_vector==0).sum()
        pos_idx = np.where(duration_vector > 1)[0]
        compensate_idx = np.random.choice(pos_idx,zero_count,replace = False)
        duration_vector[compensate_idx] -=1
        duration_vector[duration_vector==0 ] +=1
        return duration_vector
    
    def _kmer2idx(self,kmer:Union[List[int],str]):
        "Map the kmer to index, e.g. AAAAA -> 00000, AAAAC -> 00001, AAAAG -> 00002, AAAAT -> 00003, ..."
        if type(kmer) == str:
            kmer = self._num_seq(kmer)
        if len(kmer)<self.k:
            raise ValueError("Except a %dmer, but got %d"%(self.k,len(kmer)))
        multi = 1
        idx = 0
        for base in kmer[::-1]:
            idx += (int(base))*multi
            multi = multi * self.N_BASE
        return idx
    
    def build_kmer_dict(self):
        kmer2idx_dict, idx2kmer = {},[]
        for i in np.arange(self.N_BASE**self.k):
            int_list = []
            for _ in np.arange(self.k):
                int_list.append(i%self.N_BASE)
                i = i//self.N_BASE
            int_list = int_list[::-1]
            kmer = ''.join([self.ba_dict[x] for x in int_list])
            idx = self._kmer2idx(int_list)
            kmer2idx_dict[kmer] = idx
            idx2kmer.append(kmer)
        return kmer2idx_dict, idx2kmer
            
    
    def _num_seq(self,sequence:str):
        return [self.ab_dict[x] for x in sequence]
    
    
def chop(arr,chunk_length,padding = True,pad_values = 0):
    read_len = len(arr)
    chunks = np.split(arr,np.arange(0,read_len,chunk_length))[1:]
    length = [len(x) for x in chunks]
    if padding:
        last_chunk = chunks[-1]
        chunks[-1]= np.pad(last_chunk,(0,chunk_length-len(last_chunk)),'constant',constant_values = (pad_values,pad_values))
    return chunks,length

def extract(args):
    iterator = fast5_iter(args.input,mode = 'r')
    read_count = {"No basecall":0, "Kmer decode failed due to low quality":0,"Succeed":0}
    extractor = Extractor(k = args.k, alphabeta = args.alphabeta)
    config = {"k":args.k,
              "alphabeta":args.alphabeta,
              "chunk_len":args.chunk_len,
              "kmer2idx_dict":extractor.kmer2idx_dict,
              "idx2kmer":extractor.idx2kmer}
    signal_chunks, durations, kmer_chunks, meta_info = [],[],[],[]
    loop_obj = tqdm(iterator)
    for read_h,signal,fast5_f,read_id in loop_obj:
        loop_obj.set_postfix_str("no entry: %d, Kmer decode failed: %d, succeed reads:%d"%(read_count["No basecall"],read_count["Kmer decode failed due to low quality"],read_count["Succeed"]))
        try:
            seg_h = read_h['Analyses/Segmentation_%s'%(args.basecall_entry)]
            ref_sig_idx = np.asarray(seg_h["Reference_corrected/ref_sig_idx"])
            ref_seq = str(np.asarray(seg_h["Reference_corrected/ref_seq"]).astype(str))
            # basecall_entry = args.basecall_entry
        except KeyError:
            if args.alternative_entry:
                try:
                    seg_h = read_h['Analyses/Segmentation_%s'%(args.alternative_entry)]
                    ref_sig_idx = np.asarray(seg_h["Reference_corrected/ref_sig_idx"])
                    ref_seq = str(np.asarray(seg_h["Reference_corrected/ref_seq"]).astype(str))
                    # basecall_entry = args.alternative_entry #The real entry is recorded here for future possible expansion of this module.
                except KeyError:
                    read_count["No basecall"]+=1
                    continue
            else:
                read_count["No basecall"]+=1
                continue
        if args.meth:
            ref_seq = ref_seq.replace("A","M")
        if args.normalization == "noise":
            signal,_,_ = norm_by_noisiest_section(signal)
        try:
            segmented_signal,kmer_seqs = extractor.kmer_decode(ref_seq, signal[::-1], ref_sig_idx)
        except ValueError:
            read_count["Kmer decode failed due to low quality"]+=1
            continue
        if args.normalization == "dwell":
            norm_signal = dwell_normalization(segmented_signal, kmer_seqs)
        elif args.normalization == "combine":
            norm_signal = combine_normalization(segmented_signal, kmer_seqs)
        elif args.normalization == "noise":
            norm_signal = segmented_signal
        else:
            raise ValueError("Normalization method can only be dwell, combine or noise.")
        curr_chunks,curr_duration = chop(norm_signal,args.chunk_len)
        curr_kmers,_ = chop(kmer_seqs,args.chunk_len,pad_values = -1)
        signal_chunks += curr_chunks
        durations += curr_duration
        kmer_chunks += curr_kmers
        meta_info += [(fast5_f,read_id)]*len(curr_chunks)
        read_count["Succeed"] += 1
        if args.max_n and (args.max_n > 0) and (len(signal_chunks)>args.max_n):
            signal_chunks = signal_chunks[:args.max_n]
            kmer_chunks = kmer_chunks[:args.max_n]
            durations = durations[:args.max_n]
            meta_info = meta_info[:args.max_n]
            break
    np.save(os.path.join(args.output,'chunks.npy'),np.asarray(signal_chunks))
    np.save(os.path.join(args.output,'kmers.npy'),np.asarray(kmer_chunks))
    np.save(os.path.join(args.output,'durations.npy'),np.asarray(durations))
    np.savetxt(os.path.join(args.output,'meta.csv'),meta_info,fmt="%s")
    with open(os.path.join(args.output,'config.toml'),'w+') as f:
        toml.dump(config,f)
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Extract kmer information for training the NRHMM.')
    parser.add_argument("-i","--input", required = True, type = str,
                        help = "The input folder contains the fast5 files.")
    parser.add_argument("-o","--output", required = True, type = str,
                        help = "The output folder.")
    parser.add_argument("--chunk_len", default = 4000, type = int,
                        help = "The length of the chunk.")
    parser.add_argument("--basecall_entry", default = "000",type = str,
                        help = "The basecall entry.")
    parser.add_argument("--alternative_entry", default = None, type = str,
                        help = "The alternative entry to try when there is no main entry.")
    parser.add_argument("-k",type = int, default = 5,
                        help = "The k of k-mer.")
    parser.add_argument("--alphabeta", type = str, default = 'ACGTM',
                        help = "Give the order of the necleotides alphabet from 0 to n")
    parser.add_argument("--normalization",type = str, default = "noise",
                        help = "The normalization method used, can be dwell, combine or noise.")
    parser.add_argument('--max_n',
                        default = None,
                        type=int,
                        help="The maximum number of the segments to be extracted")
    parser.add_argument('--methylated',action = "store_true", dest = "meth",
                        help="This is a fully or high-proportion methylated dataset, so we replace all A with M.")
    args = parser.parse_args(sys.argv[1:])
    os.makedirs(args.output,exist_ok=True)
    extract(args)

# if __name__ == "__main__":
#     e = Extractor(5, 'ACGT')
#     seq = 'AAACGTCGTGTTT'
#     sig_start = 13
#     segemented_signal,kmer_seq = e.kmer_decode(seq, signal = np.random.rand(200), seq_pos = np.asarray([sig_start+5*x+x**2 for x in np.arange(len(seq))]))



"""
Created on Mon Aug 30 15:27:21 2021

@author: Haotian Teng
"""
import os
from typing import List
import seaborn as sns
import numpy as np
from matplotlib import pyplot as plt

def watch_errors(fs):
    errors = []
    for f in fs:
        error_file = os.path.join(f,"errors.csv")
        e = []
        with open(error_file,'r') as f:
            for line in f:
                try:
                    error = np.float(line.strip())
                except:
                    print(line.strip())
                if not np.isnan(error):
                    e.append(error)
        errors.append(e)
    return errors

def smooth(scalars: List[float], weight: float) -> List[float]:  # Weight between 0 and 1
    last = scalars[0]  # First value in the plot (first timestep)
    smoothed = list()
    for point in scalars:
        smoothed_val = last * weight + (1 - weight) * point  # Calculate smoothed value
        smoothed.append(smoothed_val)                        # Save it
        last = smoothed_val                                  # Anchor the last smoothed value
        
    return smoothed

def std_plot(errors:list,axs,**plot_args):
    ls = [len(x) for x in errors]
    max_len = max(ls)
    errors = np.asarray([np.pad(x,(0,max_len-y),'constant',constant_values = np.nan) for x,y in zip(errors,ls)])
    e_mean = np.nanmean(errors,axis = 0)
    e_std =  np.nanstd(errors,axis = 0)
    x = np.arange(len(e_mean))
    e_mean_smoothed = smooth(e_mean,plot_args['smooth'])
    axs.plot(x,e_mean_smoothed,c = plot_args['color'],linewidth = 1)
    axs.fill_between(x,e_mean - e_std,e_mean + e_std,facecolor = plot_args['color'],alpha = plot_args['alpha'],label = plot_args['label'])
    return axs

if __name__ == "__main__":
    ### Plot multiple optimizers training result
    # prefix = "/home/heavens/bridge_scratch/Xron_models_control/xron_model_supervised_control_dataset_%d_%s_16G"
    # adam_fs = [prefix%(x,'Adam') for x in np.arange(4)]    
    # sgd_fs = [prefix%(x,'SGD') for x in np.arange(4)]    
    # momentum_fs = [prefix%(x,'Momentum') for x in np.arange(4,8,1)]
    # adagrad_fs = [prefix%(x,'Adagrad') for x in np.arange(4,8,1)]    
    # colors = ['r','g','b','yellow']
    # opts = ['Adam','SGD','Momentum','Adagrad']
    # fs = [adam_fs,sgd_fs,momentum_fs,adagrad_fs]
    
    # Plot single training error
    # opts = ['2000-LN-Adam','2000-LN-Adagrad','4000-LN-Adam_1,3','4000-LN-Adam_0,2','4000-LN-Adagrad','2000-BN-Adam','2000-BN-Adagrad']
    # opts = [str(x) for x in range(8)]
    # opts = [str(x) for x in range(4)]
    colors = sns.color_palette()
    
    ##Control plotting
    
    # fs = [['/home/heavens/bridge_scratch/Xron_models_merge_d/xron_model_supervised_merge_Adagrad_transfer_learning_fixD'],
    #       ['/home/heavens/bridge_scratch/Xron_models_merge_d/xron_model_supervised_merge_Adagrad_transfer_learning_deviationCorrected'],
    #        ['/home/heavens/bridge_scratch/Xron_models_DirectTrainOnMerge/xron_model_supervised_control_dataset_%d_Adagrad_16G'%(x) for x in np.arange(8)]]
    # fs = [['/home/heavens/bridge_scratch/Xron_models_merge_d/xron_model_supervised_merge_Adagrad_transfer_learning_deviationCorrected']]
    
    # ### Plot all
    # title = "Control model accuracy"
    # repeat = 4
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_2000L/xron_model_supervised_control_dataset_%d_Adam_16G'%(x) for x in np.arange(repeat)],
    #       ['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_2000L/xron_model_supervised_control_dataset_%d_Adagrad_16G'%(x+4) for x in np.arange(repeat)],
    #       ['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_1_Adam_16G',
    #         '/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_3_Adam_16G'],
    #       ['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_0_Adam_16G',
    #         '/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_2_Adam_16G'],
    #       ['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_%d_Adagrad_16G'%(x+4) for x in np.arange(repeat)],
    #       ['/home/heavens/bridge_scratch/Xron_models_NAIVT/xron_model_supervised_control_dataset_%d_Adam_16G'%(x) for x in np.arange(repeat)],
    #       ['/home/heavens/bridge_scratch/Xron_models_NAIVT/xron_model_supervised_control_dataset_%d_Adagrad_16G'%(x+4) for x in np.arange(repeat)]]
    # opts = ['2000L-Adam'] + ['2000L-Adagrad'] + ['4000L-Adam'] + ['4000L-Adagrad'] + ['Adam'] + ['Adagrad']
    
    # ## Plot 8000L
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_8000L/xron_model_supervised_control_dataset_%d_16G'%(x)] for x in np.arange(8)]
    # opts = np.arange(8)
    # title = "8000L training accuracy"
    
    # ## Plot 4000L
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_%d_Adam_16G'%(x)] for x in np.arange(4)]
    # fs += [['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_%d_Adagrad_16G'%(x+4)] for x in np.arange(4)]
    # opts = ['Adam']*4 + ['Adagrad']*4
    
    # title = "4000L control dataset training curve"
    
    ## Plot 
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_2000L/xron_model_supervised_control_dataset_%d_Adam_16G'%(x)] for x in np.arange(4)]
    # fs += [['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_2000L/xron_model_supervised_control_dataset_%d_Adagrad_16G'%(x+4)] for x in np.arange(4)]
    # opts = ['Adam']*4 + ['Adagrad']*4
    # title = "2000L control dataset training curve"
    
    ## Plot attention model
    repeat = 4
    fs =[['/home/heavens/bridge_scratch/Xron_models_attention_NAIVT/'],
         ['/home/heavens/bridge_scratch/Xron_models_NAIVT_LayerNorm_4000L/xron_model_supervised_control_dataset_1_Adam_16G']]
    opts = ['Attention'] + ['Adam']
    title = "Attention model"
    
    ### Retrain on 100 Methylation
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT+100METH_LayerNorm_4000L/xron_model_%d_CM%d'%(x,x%4+1)] for x in np.arange(8)]
    # opts = ['Control-Methylation-Ratio %d:1, Adam'%(x+1) for x in np.arange(4)] + ['Control-Methylation-Ratio %d:1, AdamW'%(x+1) for x in np.arange(4)]
    # title = "Retrain only the last two layers"
    ###
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT+100METH_LayerNorm_4000L_retrain_all/xron_model_%d_CM%d'%(x,x%4+1)] for x in np.arange(8)]
    # opts = ['Control-Methylation-Ratio %d:1, Adam'%(x+1) for x in np.arange(4)] + ['Control-Methylation-Ratio %d:1, AdamW'%(x+1) for x in np.arange(4)]
    # title = "Retrain the whole NN."

    ###
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT+100METH+90METH_LayerNorm_4000L/xron_model_%d_CM%d'%(x,x//2%2)] for x in np.arange(8)]
    # lrs = ['1e-4','4e-5']
    # opts = ['Control-Methylation-Ratio %d:1, LR:%s'%(x//2%2,lrs[x%2]) for x in np.arange(8)]
    # title = "Retrain only the last two layers"
    ###
    # fs = [['/home/heavens/bridge_scratch/Xron_models_NAIVT+100METH+90METH_LayerNorm_4000L_retrain_all/xron_model_%d_CM%d'%(x,x//2%2)] for x in np.arange(8)]
    # lrs = ['1e-4','4e-5']
    # opts = ['Control-Methylation-Ratio %d:1, LR:%s'%(x//2%2,lrs[x%2]) for x in np.arange(8)]
    # title = "Retrain the whole NN."

    
    
    fig,axs = plt.subplots(figsize = (40,20))
    FONTSIZE = 30
    for f,c,opt in zip(fs,colors,opts):
        errors = watch_errors(f)
        std_plot(errors,axs,color = c,label = opt, alpha = .2, smooth = 0.7)
    plt.legend()
    axs.set_xlabel("Training step",fontsize = FONTSIZE)
    axs.set_ylabel("Validate Error (Editdistance/Sequence Length)",fontsize = FONTSIZE)
    plt.legend(fontsize = FONTSIZE)
    axs.set_ylim([0,1.0])
    plt.xticks(fontsize = FONTSIZE)
    plt.yticks(fontsize = FONTSIZE)
    plt.title(title,fontsize = FONTSIZE)
    prefix = "/home/heavens/bridge_scratch/Xron_models_merge_d/xron_model_supervised_control_dataset_Adagrad_16G"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  9 07:09:45 2022

@author: heavens
"""
import os
import csv
import toml
import itertools
from tqdm import tqdm
import numpy as np


class Profiler(object):
    def __init__(self,data_f, config):
        self.data_f = data_f
        print("Loading chunk data")
        chunks_f = os.path.join(data_f,"chunks.npy")
        self.chunks = np.load(chunks_f)
        self.chunk_flatten = self.chunks.flatten()
        print("Loading decoded path data")
        decoded_f = os.path.join(data_f,"path.npy")
        self.decoded = np.load(decoded_f)
        self.decoded_flatten = self.decoded.flatten()
        self.config = config
        self.idx2kmer = config['idx2kmer']
        self.kmer2idx = config['kmer2idx_dict']
        self.n_kmer = len(self.idx2kmer)
        self.alphabeta = config['alphabeta']
        self.n_base = len(self.alphabeta)
        self.transitions_count = [[[]for _ in np.arange(self.n_base)] for _ in np.arange(self.n_kmer)]
        self.transition_weight = np.zeros((self.n_kmer,self.n_base))
        self.epsilon = 1e-6
        assert np.equal(self.chunks.shape,self.decoded.shape).all()
        group_f = os.path.join(self.data_f,"grouped.npy")
        if os.path.isfile(group_f):
            print("Loading grouped array.")
            self.kmer_groups = np.load(group_f,allow_pickle = True)
        else:
            self.kmer_groups = None

    def build_invariant_kmers(self, variant_bases):
        self.invariant_kmers = []
        for i in np.arange(self.n_kmer):
            if not any([x in self.idx2kmer[i] for x in variant_bases]):
                self.invariant_kmers.append(i)
            
    def grouping(self):
        print("Sorting decoded path")
        argsort = np.argsort(self.decoded_flatten)
        sorted_chunks = self.chunk_flatten[argsort]
        print("Grouping signal")
        kmers,idxs = np.unique(self.decoded_flatten[argsort],return_index = True)
        grouped = np.split(sorted_chunks,idxs[1:])
        self.kmer_groups = [[] for _ in np.arange(self.n_kmer)]     
        for i,kmer in enumerate(kmers):
            self.kmer_groups[kmer] = grouped[i]
        print("Writing summary")
        np.save(os.path.join(self.data_f,"grouped.npy"),self.kmer_groups)

    def summarize(self):
        if self.kmer_groups is None:
            raise ValueError("Grouping file has not been found, please run grouping first.")
        self.means = np.asarray([np.mean(self.kmer_groups[i]) for i in tqdm(np.arange(self.n_kmer))])
        self.stds = np.asarray([np.std(self.kmer_groups[i]) for i in tqdm(np.arange(self.n_kmer))])
        
    def summarize_length(self):
        flatten_path = self.decoded.flatten()
        self.kmer_length = {}
        for i,kmer in tqdm(enumerate(self.idx2kmer)):
            condition = flatten_path == i
            self.kmer_length[kmer] = np.diff(np.where(np.concatenate(([condition[0]],
                                                                      condition[:-1] != condition[1:],
                                                                      [True])))[0])[::2]
    
    def summarize_transition(self):
        for path in self.decoded:
            keys,counts = [],[]
            for key,group in itertools.groupby(path):
                keys.append(key)
                counts.append(len(list(group)))
            for i,k in enumerate(keys):
                if i == len(keys) - 1:
                    break
                next_base = self.idx2kmer[keys[i+1]][-1]
                self.transitions_count[k][self.alphabeta.index(next_base)].append(counts[i])
        for i in np.arange(self.n_kmer):
            for j in np.arange(self.n_base):
                c = self.transitions_count[i][j]
                self.transition_weight[i][j] = len(c)/(sum(c)+self.epsilon) #ML estimation of transition probability is N/sum_i k_i - N respect that the stay probability is 1, see the notebook for more details.
    
if __name__ == "__main__":
    home_f = os.path.expanduser("~")
    control_folder = home_f + "/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_guppy_4000_noise"
    m6a_folder = home_f + "/bridge_scratch/ime4_Yearst/IVT/m6A/rep2/kmers_guppy_4000_noise"
    config = toml.load(os.path.join(control_folder,"config.toml"))
    p_control= Profiler(control_folder,config)
    # p_control.summarize()
    # p_control.summarize_length()
    p_control.summarize_transition()
    p_m6A= Profiler(m6a_folder,config)
    # p_m6A.summarize()
    # p_m6A.summarize_length()
    p_m6A.summarize_transition()
    
        


"""
Created on Fri Jun 17 23:00:01 2022

@author: Haotian Teng
"""
import os 
import sys
import toml
import random
import argparse
import numpy as np
from typing import List
import itertools
from tqdm import tqdm
from xron.utils.seq_op import Methylation_DP_Aligner
def kmers2seq(kmers,idx2kmer,return_pos = False):
    merged = [g for g,_ in itertools.groupby(kmers)]
    seqs = [idx2kmer[x][0] for x in merged]
    if return_pos:
        pos = [len(list(G)) for _,G in itertools.groupby(kmers)]
        return ''.join(seqs) + idx2kmer[merged[-1]][1:],pos
    return ''.join(seqs) + idx2kmer[merged[-1]][1:]

def transfer(args):
    kmers = np.load(os.path.join(args.input,"kmers.npy"),mmap_mode = "r")
    durations = np.load(os.path.join(args.input,"durations.npy"),mmap_mode = "r")
    config = toml.load(os.path.join(args.input,"config.toml"))
    seqs,seq_lens = [],[]
    seqs = [kmers2seq(x[:durations[i]],config['idx2kmer']) for i,x in tqdm(enumerate(kmers))]
    seq_lens = [len(x) for x in seqs]
    np.save(os.path.join(args.input,'seqs.npy'),seqs)
    np.save(os.path.join(args.input,'seq_lens.npy'),seq_lens)

class LinkingAlignment(object):
    def __init__(self,aln_seq,duration,aln_ref):
        """Linking the alignment sequence and reference sequence together.
        duration:    d0 d1 0  0  d2 d3 d4 d5
                     |  |  |  |  |  |  |  |
        seq          A->C->_->_->G->A->C->T
                     |  |  |  |  |  |  |  |
        ref          A->_->C->T->_->_->C->T
        """
        self.root = [SeqNode('$',0),RefNode('$')]
        j = 0 #The index of the duration
        #Linking the sequence
        prev = self.root[0]
        for i,c in enumerate(aln_seq):
            curr_d = duration[j] if c !='_' else 0
            if c != '_':
                j+=1
            curr = SeqNode(c,curr_d)
            curr.prev = prev
            prev.next = curr
            prev = curr
        #Linking the reference
        prev = self.root[1]
        for i,c in enumerate(aln_ref):
            curr = RefNode(c)
            curr.prev = prev
            prev.next = curr
            prev = curr
        #Linking the sequence and reference
        curr_q = self.root[0]
        curr_r = self.root[1]
        while curr_q is not None:
            curr_q.ref = curr_r
            curr_r.q = curr_q
            curr_q = curr_q.next
            curr_r = curr_r.next
    
    @property
    def seq(self):
        curr = self.root[0].next
        seq = ''
        while curr is not None:
            seq += curr.c
            curr = curr.next
        return seq
    
    @property
    def ref(self):
        curr = self.root[1].next
        seq = ''
        while curr is not None:
            seq += curr.c
            curr = curr.next
        return seq
    
    @property
    def durations(self):
        curr = self.root[0].next
        durations = []
        while curr is not None:
            durations.append(curr.duration)
            curr = curr.next
        return durations

    def insertion_traverse_fix(self):
        #Fix the insertion in the sequence.
        curr = self.root[0].next
        while curr is not None:
            if curr.ref.c == '_':
                curr = self._carve_up_even(curr)
            else:
                curr = curr.next
    
    def deletion_traverse_fix(self):
        #Fix the deletion in the reference.
        curr = self.root[1].next
        while curr is not None:
            if curr.q.c == '_':
                curr = self._apportion_even(curr.q)
                curr = curr.ref if curr is not None else None
            else:
                curr = curr.next

    def mismatch_traverse_fix(self):
        #Fix the mismatch in the sequence.
        curr = self.root[0].next
        while curr is not None:
            if curr.c != curr.ref.c and curr.c != '_' and curr.ref.c != '_':
                curr.c = curr.ref.c
            curr = curr.next
    
    def _carve_up_left(self,seq_node):
        """Carve up chosen element into its left node, correction for insertion in the sequence.
        """
        if seq_node.prev is None or seq_node.prev.c == '$':
            #There is no prefix node
            seq_node.next.duration += seq_node.duration
            seq_node.prev.next = seq_node.next
            seq_node.next.prev = seq_node.prev
            seq_node.ref.prev.next = seq_node.ref.next
            seq_node.ref.next.prev = seq_node.ref.prev
        else:
            #There is no suffix node
            seq_node.prev.duration += seq_node.duration
            seq_node.prev.next = seq_node.next
            seq_node.next.prev = seq_node.prev
            seq_node.ref.prev.next = seq_node.ref.next
            seq_node.ref.next.prev = seq_node.ref.prev
        del seq_node.ref
        curr_node = seq_node.next
        del seq_node
        return curr_node

    def _carve_up_right(self,seq_node):
        """Carve up chosen element into its left node, correction for insertion in the sequence.
        """
        if seq_node.next is None:
            #There is no suffix node
            seq_node.prev.duration += seq_node.duration
            seq_node.prev.next = seq_node.next
            seq_node.next.prev = seq_node.prev
            seq_node.ref.prev.next = seq_node.ref.next
            seq_node.ref.next.prev = seq_node.ref.prev
        else:
            #carve to suffix node.
            seq_node.next.duration += seq_node.duration
            seq_node.prev.next = seq_node.next
            seq_node.next.prev = seq_node.prev
            seq_node.ref.prev.next = seq_node.ref.next
            seq_node.ref.next.prev = seq_node.ref.prev
        del seq_node.ref
        curr_node = seq_node.next
        del seq_node
        return curr_node

    def _carve_up_even(self,seq_node):
        """Carve up chosen element into its neighbourhood evenly, correction for insertion in the sequence.
        """
        if seq_node.prev is None or seq_node.prev.c == '$':
            #There is no prefix node
            seq_node.next.duration += seq_node.duration
            if seq_node.prev:
                seq_node.prev.next = seq_node.next
                seq_node.ref.prev.next = seq_node.ref.next
            seq_node.next.prev = seq_node.prev
            seq_node.ref.next.prev = seq_node.ref.prev
        elif seq_node.next is None:
            #There is no suffix node
            seq_node.prev.duration += seq_node.duration
            seq_node.prev.next = seq_node.next
            seq_node.ref.prev.next = seq_node.ref.next
        else:
            #Otherwise divide the current node evenly
            seq_node.prev.duration += seq_node.duration//2
            seq_node.next.duration += seq_node.duration//2
            random.choice([seq_node.prev,seq_node.next]).duration += seq_node.duration%2
            seq_node.prev.next = seq_node.next
            seq_node.next.prev = seq_node.prev
            seq_node.ref.prev.next = seq_node.ref.next
            seq_node.ref.next.prev = seq_node.ref.prev
        del seq_node.ref
        curr_node = seq_node.next
        del seq_node
        return curr_node

    def _apportion_even(self,seq_node):
        """Apportion duration from nearby nodes, correction for deletion in the sequence."""
        seq_node.c = seq_node.ref.c
        if seq_node.prev.c != '$':
            #Borrow from the previous node
            share_node = seq_node.prev
            share_ratio = 3
            while share_node is not None and share_node.duration < 2:
                share_node = share_node.prev
                share_ratio += 1
            if share_node is not None:
                share = max(1,int(share_node.duration//share_ratio))
                seq_node.duration += share
                share_node.duration -= share
        else:
            #The insertion is happened at the beginning of the sequence, delete it.
            seq_node.prev.next = seq_node.next
            seq_node.next.prev = seq_node.prev
            seq_node.ref.prev.next = seq_node.ref.next
            seq_node.ref.next.prev = seq_node.ref.prev
            del seq_node.ref
            curr_node = seq_node.next
            del seq_node
            return curr_node

        if seq_node.next is not None:
            #Divide duration from the next node
            share_node = seq_node.next
            share_ratio = 3
            while share_node is not None and share_node.duration < 2:
                share_node = share_node.next
                share_ratio += 1
            if share_node is None:
                #There is no next node to borrow from, skip it
                if seq_node.duration > 0:
                    seq_node.c = seq_node.ref.c
                    return seq_node.next
                else:
                    #The node can't borrow duration from anywhere. delete it.
                    seq_node.prev.next = seq_node.next
                    seq_node.next.prev = seq_node.prev
                    seq_node.ref.prev.next = seq_node.ref.next
                    seq_node.ref.next.prev = seq_node.ref.prev
                    del seq_node.ref
                    curr_node = seq_node.next
                    del seq_node
                    return curr_node
            share = max(1,int(share_node.duration//share_ratio))
            seq_node.duration += share
            share_node.duration -= share
        seq_node.c = seq_node.ref.c
        return seq_node.next

class SeqNode(object):
    def __init__(self,c,duration):
        self.c = c
        self.duration = duration
        self.next = None
        self.prev = None
        self.ref = None #Linking to a RefNode

class RefNode(object):
    def __init__(self,c):
        self.c = c
        self.q = None #Linking to a SeqNode
        self.next = None
        self.prev = None

def fixing_looping_path(path,seq,idx2kmer:List,modified_base,canonical_base):
    """Fixing the looping path like AACCAACCAACCA to AACCA due to circle on HMM.
    """
    kmer_n = len(idx2kmer[0])
    decoded,duration = kmers2seq(path,idx2kmer,return_pos = True)
    if decoded == seq:
        return path,False
    duration += [0]*(kmer_n-1)
    aligner = Methylation_DP_Aligner(base_alternation = {modified_base:canonical_base})
    x,y = aligner.align(decoded, seq)
    aln = LinkingAlignment(x,duration,y)
    aln.insertion_traverse_fix()
    aln.deletion_traverse_fix()
    aln.mismatch_traverse_fix()
    fix_seq,fix_duration = aln.seq,aln.durations
    if len(fix_seq) <= kmer_n:
        return path,None #The path is too short to be fixed
    fix_path = seq2path(fix_seq,fix_duration,idx2kmer)
    return np.asarray(fix_path),True

def seq2path(seq,duration,idx2kmer:List):
    """Convert a sequence to a path.
    """
    kmer_n = len(idx2kmer[0])
    path = []
    for i in range(0,len(seq)-kmer_n):
        kmer = seq[i:i+kmer_n]
        path += [idx2kmer.index(kmer)]*duration[i]
    path += [idx2kmer.index(seq[-kmer_n:])]*(sum(duration[-kmer_n:]))
    return path

if __name__ == "__main__":
    # parser = argparse.ArgumentParser(
    #     description='Transfer the kmer-seq to sequences.')
    # parser.add_argument('-i', '--input', required = True,
    #                     help = "The folder contains the kmers.npy and durations.npy files.")
    # args = parser.parse_args(sys.argv[1:])
    # transfer(args)
    config = toml.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_guppy_4000_noise/config.toml")
    seqs = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_guppy_4000_noise/seqs.npy")
    paths = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_guppy_4000_noise/path.npy")
    fix_path = fixing_looping_path(paths[11],seqs[11],idx2kmer = config['idx2kmer'],canonical_base = 'A',modified_base = 'M')


# This script is written to include the operations including the coordinates mapping and transfer.
def reverse_mapping(t2g_dict):
    """Get the g2t_dict from t2g_dict, mapping genomics position to transcript position, 
    note that each genomics position can be mapped to multiple transcript positions.
    param t2g_dict[Dict]: transcript to genomics position dictionary
        key: transcript id
        value: a dict containing the mapping from transcript position to genomics position
            key: transcript position
            value: a tuple of (tx_contig, g_id, g_pos)
    return g2t_dict[Dict]: genomics to transcript position dictionary
        key: genomics contig
        value: a dictionary of genomics position 
            key: genomics position
            value: a list of tuple [(t1_id, t1_pos),(t2_id, t2_pos),...] containing all the possible transcript positions
    """
    g2t_dict = {}
    for t_id,t in t2g_dict.items():
        for t_pos, g in t.items():
            if g[0] not in g2t_dict:
                g2t_dict[g[0]] = {}
            if g[2] not in g2t_dict[g[0]]:
                g2t_dict[g[0]][g[2]] = []
            g2t_dict[g[0]][g[2]].append((t_id, t_pos))
    return g2t_dict

def t2g(tx_id, fasta_dict, gtf_dict):
    t2g_dict = {}
    if tx_id not in fasta_dict.keys():
        return [], []
    tx_seq = fasta_dict[tx_id]
    tx_contig = gtf_dict[tx_id]['chr']
    g_id = gtf_dict[tx_id]['g_id']
    if tx_seq is None:
        return [], []

    for exon_num in range(len(gtf_dict[tx_id]['exon'])):
        g_interval = gtf_dict[tx_id]['exon'][exon_num]
        tx_interval = gtf_dict[tx_id]['tx_exon'][exon_num]
        for g_pos in range(g_interval[0], g_interval[1] + 1): # Exclude the rims of exons.
            dis_from_start = g_pos - g_interval[0]
            if gtf_dict[tx_id]['strand'] == "+":
                tx_pos = tx_interval[0] + dis_from_start
            elif gtf_dict[tx_id]['strand'] == "-":
                tx_pos = tx_interval[1] - dis_from_start
            if (g_interval[0] <= g_pos < g_interval[0]+2) or (g_interval[1]-2 < g_pos <= g_interval[1]): 
                kmer = 'XXXXX'
            else:
                kmer = tx_seq[tx_pos-2:tx_pos+3]
            t2g_dict[tx_pos] = (tx_contig, g_id, g_pos) # tx.contig is chromosome.
    return t2g_dict

def readFasta(transcript_fasta_paths_or_urls):
    fasta=open(transcript_fasta_paths_or_urls,"r")
    entries=""
    for ln in fasta:
        entries+=ln
    entries=entries.split(">")
    dict={}
    for entry in entries:
        entry=entry.split("\n")

        if len(entry[0].split()) > 0:
            id=entry[0].split()[0]
            seq="".join(entry[1:])
            dict[id]=seq
    return dict

def readGTF(gtf_path_or_url):
    gtf=open(gtf_path_or_url,"r")
    dict={}
    gene_transcript_dict={}
    for ln in gtf:
        if not ln.startswith("#"):
            ln=ln.strip("\n").split("\t")
            if ln[2] == "transcript" or ln[2] in ("exon", "start_codon", "stop_codon", "CDS", 
                                                  "three_prime_utr", "five_prime_utr"):
                chr,type,start,end=ln[0],ln[2],int(ln[3]),int(ln[4])
                attrList=ln[-1].split(";")
                attrDict={}
                for k in attrList:
                    p=k.strip().split(" ")
                    if len(p) == 2:
                        attrDict[p[0]]=p[1].strip('\"')
                tx_id = attrDict["transcript_id"] + "." + attrDict["transcript_version"]
                g_id = attrDict["gene_id"]
                if g_id not in gene_transcript_dict:
                    gene_transcript_dict[g_id] = [tx_id]
                else:
                    gene_transcript_dict[g_id].append(tx_id)
                if tx_id not in dict:
                    dict[tx_id]={'chr':chr,'g_id':g_id,'strand':ln[6]}
                    if type not in dict[tx_id]:
                        if type == "transcript":
                            dict[tx_id][type]=(start,end)
                else:
                    if type == 'CDS':
                        info = (start, end, int(attrDict['exon_number']))
                    else:
                        info = (start, end)
                    if type not in dict[tx_id]:
                        dict[tx_id][type]=[info]
                    else:
                        dict[tx_id][type].append(info)
                          
    #convert genomic positions to tx positions
    for id in dict:
        tx_pos,tx_start=[],0
        for pair in dict[id]["exon"]:
            tx_end=pair[1]-pair[0]+tx_start
            tx_pos.append((tx_start,tx_end))
            tx_start=tx_end+1
        dict[id]['tx_exon']=tx_pos
    return dict,gene_transcript_dict

def map_nm_to_enst(nm_id):
    # Step 1: Retrieve the RefSeq accession associated with the NM ID using NCBI E-utilities API
    ncbi_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"
    ncbi_params = {
        "db": "nucleotide",
        "id": nm_id,
        "rettype": "acc",
        "retmode": "text"
    }
    ncbi_response = requests.get(ncbi_url, params=ncbi_params)
    refseq_acc = ncbi_response.text.strip()

    # Step 2: Map the RefSeq accession to an Ensembl transcript ID using Ensembl REST API
    ensembl_url = "https://rest.ensembl.org/map/cdna/" + refseq_acc + "/GRCh38"
    ensembl_headers = {"Content-Type": "application/json"}
    ensembl_response = requests.get(ensembl_url, headers=ensembl_headers)
    ensembl_json = ensembl_response.json()
    if "mappings" in ensembl_json and len(ensembl_json["mappings"]) > 0:
        ensembl_id = ensembl_json["mappings"][0]["to_id"]
        return ensembl_id
    else:
        return None
    
if __name__ == "__main__":
    #%% Set up paths
    import os
    import pandas as pd
    import pickle
    from itertools import chain
    SCRATCH = os.environ['SCRATCH']
    ref_folder = os.path.join(SCRATCH, 'NA12878_RNA_IVT/GRCh38_transcript_ensembel_sgnex')
    trascript_f = os.path.join(ref_folder,'Homo_sapiens.GRCh38.cdna.ncrna.fa')
    gtf_f = os.path.join(ref_folder,'Homo_sapiens.GRCh38.91.gtf')
    sites_f = f"{SCRATCH}/Xron_Project/m6A_site_m6Anet_DRACH_HEK293T.csv"

    #%%Test case 1:
    #Build t2g dictionary for human reference from SG-NEX data
    out_f = f"{SCRATCH}/Xron_Project/Benchmark/HEK293T/t2g_dict.pkl"
    print("Reading reference fasta files...")   
    fasta = readFasta(trascript_f)
    print("Reading reference gtf files...")
    gtf,id_map = readGTF(gtf_f)
    print("Reading transcript names...")
    df = pd.read_csv(sites_f)
    gene_ids = df['gene_id'].unique()
    tx_names = [id_map[gene_id] for gene_id in gene_ids if gene_id in id_map]
    tx_names = list(chain(*tx_names)) #flatten the list
    print("Creating t2g dictionary...")
    t2g_dict = {tx:t2g(tx, fasta, gtf) for tx in tx_names}
    print("Saving t2g dictionary...")
    with open(out_f, "wb+") as f:
        pickle.dump(t2g_dict, f)

    #%%Test case 2:
    #Build reverse g2t dictionary for human reference 
    t2g_f = f"{SCRATCH}/Xron_Project/Benchmark/HEK293T/t2g_dict.pkl"
    out_f = f"{SCRATCH}/Xron_Project/Benchmark/HEK293T/g2t_dict.pkl"
    print("Reading the t2g dictionary...")
    if not os.path.exists(t2g_f):
        print("t2g dictionary not found. Please run the first test case first.")
        exit(1)
    with open(t2g_f, "rb") as f:
        t2g_dict = pickle.load(f)
    print("Creating g2t dictionary...")
    g2t_dict = reverse_mapping(t2g_dict)
    print("Saving g2t dictionary...")
    with open(out_f, "wb+") as f:
        pickle.dump(g2t_dict, f)
    
# %%



"""
Created on Wed Jul 20 11:04:29 2022

@author: Haotian Teng
"""
import numpy as np
from xron.utils.fastIO import read_fast5
from xron.utils.seq_op import fast5_iter
from itertools import groupby
from tqdm import tqdm
import argparse
import sys

def softmax(logits,axis = -1):
    return np.exp(logits)/np.sum(np.exp(logits),axis = axis,keepdims = True)

def get_posterior(logits,seq,move,canonical_base = 'A',modified_base = 'M',base_list = ['A','C','G','T','M'],n_largest_p = 3):
    """Get the posterior probability of the canonical base and modified base"""
    pos = np.cumsum(move)-1
    posterior = []
    for k,g in groupby(zip(logits,pos),lambda x:x[1]):
        g = np.asarray([x[0] for x in g])
        if k >= len(seq):
            print("Warning, found sequence that is too short, probabily the result is from a overlay>0 basecall.")
            return None
        if seq[k] == canonical_base or seq[k] == modified_base:
            g = softmax(g)
            p_canonical = g[:,base_list.index(canonical_base)+1]
            p_canonical.sort()
            p_canonical = p_canonical[-n_largest_p:].sum()
            p_modified = g[:,base_list.index(modified_base)+1]
            p_modified.sort()
            p_modified = p_modified[-n_largest_p:].sum()
            posterior.append(p_modified/(p_canonical+p_modified))
    return np.asarray(posterior)

def write_modified_probability(args):
    fail_count = 0
    with tqdm() as t:
        for read_h,signal,abs_path,read_id in fast5_iter(args.fast5,mode = 'a'):
            t.postfix = "File: %s, Read: %s, failed: %d"%(abs_path,read_id,fail_count)
            try:
                logits,move,seq = read_fast5(read_h,index = args.basecall_entry)
            except:
                fail_count += 1
                continue
            mod_p = get_posterior(logits,
                                  seq,
                                  move,
                                  canonical_base = args.canonical_base,
                                  modified_base = args.modified_base,
                                  base_list = [x for x in args.alphabeta],
                                  n_largest_p = args.n_largest_p)
            if mod_p is None:
                fail_count += 1
                continue
            result_h = read_h['Analyses/Basecall_1D_%s/BaseCalled_template'%(args.basecall_entry)]
            if 'ModifiedProbability' in result_h:
                del result_h['ModifiedProbability']
            result_h.create_dataset('ModifiedProbability',data = mod_p,dtype = "f")
            result_h['ModifiedProbability'].attrs['alphabet'] = args.alphabeta
            result_h['ModifiedProbability'].attrs['canonical_base'] = args.canonical_base
            result_h['ModifiedProbability'].attrs['modified_base'] = args.modified_base
            result_h['ModifiedProbability'].attrs['n_largest_p'] = args.n_largest_p
            t.update()
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-i','--fast5',type = str,required = True,help = 'Path to the fast5 folder')
    parser.add_argument('--basecall_entry',type = str,default = '000',help = 'Basecall entry')
    parser.add_argument('--alphabeta',type = str,default = 'ACGTM',help = 'Alphabet of the basecall model. Default is ACGTM')
    parser.add_argument('--canonical_base',type = str,default = 'A',help = 'Canonical base. Default is A')
    parser.add_argument('--modified_base',type = str,default = 'M',help = 'Modified base. Default is M')
    parser.add_argument('--n_largest_p',type = int,default = 3,help = 'Number of largest probability to use. Default is 3')
    args = parser.parse_args(sys.argv[1:])
    write_modified_probability(args)


"""
Created on Fri May 20 14:53:39 2022

@author: Haotian Teng
"""

import os
import sys
import toml
import torch
import argparse
import itertools
import numpy as np
from tqdm import tqdm
from typing import Dict,List
from xron.utils.seq_op import Methylation_DP_Aligner
from xron.nrhmm.hmm import GaussianEmissions, RHMM
from xron.nrhmm.hmm_input import Kmer2Transition, Kmer_Dataset, Normalizer
from torchvision import transforms
from torch.utils.data.dataloader import DataLoader
from xron.xron_train_base import DeviceDataLoader

def kmers2seq(kmers,idx2kmer):
    merged = [g for g, _ in itertools.groupby(kmers)]
    seqs = [idx2kmer[x][0] for x in merged]
    return ''.join(seqs) + idx2kmer[merged[-1]][1:]

def print_pore_model(emission_means,idx2kmer,f = None):
    for i,mean in enumerate(emission_means):
        if f:
            f.write("%s: %.2f\n"%(idx2kmer[i],mean))
        else:
            print("%s: %.2f"%(idx2kmer[i],mean))

def get_effective_kmers(selection:str,idx2kmer:List):
    demand_char = [x for x in selection.split("!")[0]]
    filter_char = [x for x in selection.split("!")[1]]
    selected_kmers = [i for i,x in enumerate(idx2kmer) if all([d in x for d in demand_char] + [f not in x for f in filter_char])]
    return selected_kmers

def load(self,save_folder,update_global_step = True):
    self.save_folder = save_folder
    ckpt_file = os.path.join(save_folder,'checkpoint')
    with open(ckpt_file,'r') as f:
        latest_ckpt = f.readline().strip().split(':')[1]
        if update_global_step:
            self.global_step = int(latest_ckpt.split('-')[1])
    ckpt = torch.load(os.path.join(save_folder,latest_ckpt),
                      map_location=self.device)
    for key,net in ckpt.items():
        if key in self.nets.keys():
            self.nets[key].load_state_dict(net,strict = False)
            self.nets[key].to(self.device)
        else:
            print("%s net is defined in the checkpoint but is not imported because it's not defined in the model."%(key))

def main(args):
    print("Loading data...")
    config = toml.load(os.path.join(args.input,"config.toml"))
    chunks = np.load(os.path.join(args.input,"chunks.npy"),mmap_mode=args.mmap_mode)
    n_samples, sig_len = chunks.shape
    durations = np.load(os.path.join(args.input,"durations.npy"),mmap_mode = args.mmap_mode)
    idx2kmer = config['idx2kmer']
    kmers = np.load(os.path.join(args.input,"kmers.npy"),mmap_mode = args.mmap_mode)
    k2t = Kmer2Transition(alphabeta = config['alphabeta'],
                          k = config['k'],
                          T_max = config['chunk_len'],
                          kmer2idx = config['kmer2idx_dict'],
                          idx2kmer = config['idx2kmer'],
                          neighbour_kmer = 4 ,
                          base_prior = {x:args.transition_prior for x in config['alphabeta']},
                          base_alternation = {"A":"M"}, 
                          kmer_replacement = args.kmer_replacement,
                          out_format = args.transition_type)
    dataset = Kmer_Dataset(chunks, durations, kmers,transform=transforms.Compose([k2t]))
    loader = DataLoader(dataset,batch_size = args.batch_size, shuffle = False)
    loader = DeviceDataLoader(loader,device = args.device)
    
    print("Load the model.")
    ckpt_file = os.path.join(args.model,'checkpoint')
    with open(ckpt_file,'r') as f:
        latest_ckpt = f.readline().strip().split(':')[1]
    model = torch.load(os.path.join(args.model,latest_ckpt),
                       map_location = torch.device(args.device))
    emission = GaussianEmissions(model['hmm']['emission.means'].cpu().numpy(), model['hmm']['emission.cov'].cpu().numpy())
    hmm = RHMM(emission,
               normalize_transition=False,
               device = args.device,
               transition_operation = args.transition_operation,
               index_mapping = k2t.idx_map if args.transition_operation == "compact" else None)
    
    print("Readout the pore model.")
    with open(os.path.join(args.model,"pore_model"), "w+") as f:
        print_pore_model(model['hmm']['emission.means'],config['idx2kmer'],f)

    aligner = Methylation_DP_Aligner(base_alternation = {'M':'A'})
    if args.effective == "!":
        effective_kmers = None
    else:
        effective_kmers = get_effective_kmers(args.effective, idx2kmer)
    norm = Normalizer(use_dwell = True,
                      effective_kmers=effective_kmers)
    sigs,seqs,paths = [],[],[]
    for i_batch, batch in tqdm(enumerate(loader)):
        renorm_batch = []
        signal_batch = batch['signal']
        duration_batch = batch['duration']
        transition_batch = batch['labels']
        kmers_batch = batch['kmers']
        with torch.no_grad():
            path,logit = hmm.viterbi_decode(signal_batch, duration_batch, transition_batch)
        paths.append(path.cpu().numpy())
        renorm_batch = signal_batch
        for j in tqdm(np.arange(args.renorm+1),desc = "Renorm the signal:"):
            with torch.no_grad():
                path,logit = hmm.viterbi_decode(renorm_batch, duration_batch, transition_batch)
            rc_signal = np.asarray([[hmm.emission.means[x].item() for x in p] for p in path.cpu().numpy()])
            renorm_batch = norm(renorm_batch.cpu().numpy()[:,:,0],rc_signal, duration_batch.cpu().numpy(),path.cpu().numpy())
            renorm_batch = torch.from_numpy(renorm_batch).unsqueeze(dim = -1).to(args.device)
        renorm_batch = renorm_batch.squeeze(dim = -1).cpu().numpy()
        sigs.append(renorm_batch)
        for i in np.arange(args.batch_size):
            if i >= signal_batch.shape[0]:
                continue
            deco_seq = kmers2seq(path.cpu().numpy()[i][:duration_batch[i]],idx2kmer)
            orig_seq = kmers2seq(kmers_batch.cpu().numpy()[i][:duration_batch[i]],idx2kmer)
            deco_aln,orig_aln = aligner.align(deco_seq,orig_seq)
            final_seq = aligner.merge(deco_aln,orig_aln)
            seqs.append(final_seq)
        if args.max_n:
            if len(seqs) >= args.max_n:
                break
    seq_lens = [len(i) for i in seqs]
    sigs = np.vstack(sigs)
    seqs = np.array(seqs)
    seq_lens = np.array(seq_lens)
    paths = np.concatenate(paths,axis = 0)
    np.save(os.path.join(args.input,'chunks_renorm.npy'),sigs[:args.max_n])
    np.save(os.path.join(args.input,'seqs_re.npy'),seqs[:args.max_n])
    np.save(os.path.join(args.input,'seq_re_lens.npy'),seq_lens[:args.max_n])
    np.save(os.path.join(args.input,'path'),paths[:args.max_n])

def add_arguments(parser):
    parser.add_argument("-i","--input", type = str, required = True,
                        help = "Data folder contains the chunk, kmer sequence.")
    parser.add_argument('-m', '--model', required = True,
                        help = "The rhmm model folder.")
    parser.add_argument("-b","--batch_size", type = int, default = 20,
                        help = "The batch size to train.")
    parser.add_argument("--renorm", type = int, default = 1,
                        help = "Number of time to renormalize the signal.")
    parser.add_argument("--device",type = str, default = None,
                        help = "The device used to train the model.")
    parser.add_argument("--max_n",type = int, default = None,
                        help = "The maximum number of reads.")
    parser.add_argument("--certain_methylation",action = "store_false", 
                        dest = "kmer_replacement", 
                        help = "If we are sure about the methylation state.")
    parser.add_argument("--mmap_mode",type = str, default = None,
                        help = "mmap mode when loding numpy data, default is\
                        None which does not enable mmapmode, can be r.")
    parser.add_argument("-e","--effective",type = str, default = "!",
                        help = "A magic string gives the kmers that take into \
                            account when doing normalization, for example, \
                            A!M means kmers that must have A and must not \
                            have M is taken into account.")
    parser.add_argument("--transition_prior",type = float, default = 0.3,
                        help = "The prior probability of transition matrix.")
    parser.add_argument("--transition_operation",type = str, default = "sparse")

def post_args(args):
    if args.device is None:
        args.device = "cuda" if torch.cuda.is_available() else "cpu"
    if args.transition_operation == "compact":
        args.transition_type = "compact"
    else:
        args.transition_type = "sparse"

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training RHMM model')
    args = parser.parse_args(sys.argv[1:])
    main(args)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Sep  3 14:35:26 2022

@author: heavens
"""
import os
import numpy as np

def read_fastqs(fastq_f):
    records = {"sequences":[],"name":[],"quality":[]}
    for fastq in os.listdir(fastq_f):
        with open(os.path.join(fastq_f,fastq),'r') as f:
            for line in f:
                if line.startswith("@"):
                    records['name'].append(line.strip()[1:])
                    records['sequences'].append(next(f).strip())
                    assert next(f).strip() == "+" #skip the "+"
                    records['quality'].append(next(f).strip())
    return records

def read_fastq(fastq):
    records = {"sequences":[],"name":[],"quality":[]}
    with open(fastq,'r') as f:
        for line in f:
            if line.startswith("@"):
                records['name'].append(line.strip()[1:])
                records['sequences'].append(next(f).strip())
                assert next(f).strip() == "+",print(line) #skip the "+"
                records['quality'].append(next(f).strip())
    return records

def read_fast5(read_h,index = "000"):
    result_h = read_h['Analyses/Basecall_1D_%s/BaseCalled_template'%(index)]
    logits = result_h['Logits']
    move = result_h['Move']
    try:
        seq = str(np.asarray(result_h['Fastq']).astype(str)).split('\n')[1]
    except:
        seq = np.asarray(result_h['Fastq']).tobytes().decode('utf-8').split('\n')[1]
    return np.asarray(logits),np.asarray(move),seq

def read_entry(read_h,entry:str,index = "000"):
    """
    Read a entry given the name

    """
    result_h = read_h['Analyses/Basecall_1D_%s/BaseCalled_template'%(index)]
    return np.asarray(result_h[entry])


"""
Created on Wed May 11 02:05:35 2022

@author: Haotian Teng
"""
import os
import time
import toml
import torch
import itertools
import numpy as np
from matplotlib import pyplot as plt
from xron.nrhmm.hmm import GaussianEmissions, RHMM
from xron.nrhmm.hmm_input import Kmer2Transition, Kmer_Dataset
from torchvision import transforms
from torch.utils.data.dataloader import DataLoader
from xron.xron_train_base import DeviceDataLoader

def kmers2seq(kmers,idx2kmer):
    merged = [g for g,_ in itertools.groupby(kmers)]
    seqs = [idx2kmer[x][2] for x in merged]
    return ''.join(seqs)

SCRATCH = os.environ['SCRATCH']
out_f = f"{SCRATCH}/Xron_Project/figures"
torch.manual_seed(1992)
model = torch.load(f"{SCRATCH}/NRHMM_models/xron_rhmm_models_new/ckpt-36234")
emission = GaussianEmissions(model['hmm']['emission.means'].cpu().numpy(), 1*np.ones(3125)[:,None])
class TestArguments:
    input = f"{SCRATCH}/NA12878_RNA_IVT/xron_partial/extracted_kmers/"
    # input = "/home/heavens/bridge_scratch/m6A_Nanopore_RNA002/data/m6A_90_pct/20210430_1745_X2_FAQ15454_23428362/kmers_guppy_4000/"
    batch_size = 10
    device = "cuda"
args = TestArguments
hmm = RHMM(emission,normalize_transition=False,device = args.device)
config = toml.load(os.path.join(args.input,"config.toml"))
print("Readout the pore model.")
chunks = np.load(os.path.join(args.input,"chunks.npy"))
n_samples, sig_len = chunks.shape
durations = np.load(os.path.join(args.input,"durations.npy"))
idx2kmer = config['idx2kmer']
kmers = np.load(os.path.join(args.input,"kmers.npy"))
base_prior = {x:1 for x in config['alphabeta']}
k2t = Kmer2Transition(alphabeta = config['alphabeta'],
                      k = config['k'],
                      T_max = config['chunk_len'],
                      kmer2idx = config['kmer2idx_dict'],
                      idx2kmer = config['idx2kmer'],
                      neighbour_kmer = 2,
                      base_alternation = {"A":"M"},
                      base_prior = base_prior,
                      kmer_replacement = True)
dataset = Kmer_Dataset(chunks, durations, kmers,transform=transforms.Compose([k2t]))
loader = DataLoader(dataset,batch_size = args.batch_size, shuffle = True)
loader = DeviceDataLoader(loader,device = args.device)
for i_batch, batch in enumerate(loader):
    signal_batch = batch['signal']
    duration_batch = batch['duration']
    transition_batch = batch['labels']
    kmers_batch = batch['kmers']
    Ls = duration_batch.cpu().numpy()
    break
idx = 6
signal = signal_batch[idx].detach().cpu().numpy()[:400]
kmers = kmers_batch[idx].detach().cpu().numpy()[:400]
seq = kmers2seq(kmers,idx2kmer)

## Print the alignment
fig,axs = plt.subplots(nrows = 2, 
                       ncols = 2, 
                       gridspec_kw={'width_ratios': [1, 10],
                                    "height_ratios":[1,5],
                                    'wspace':0, 
                                    'hspace':0},
                       figsize = (6,6))
for i, ax in enumerate(fig.axes):
    ax.set_xticklabels([])
    ax.set_yticklabels([])
axs[0][0].set_axis_off()
axs[0][1].plot(signal)
axs[0][1].xaxis.set_visible(False)
for i,c in enumerate(seq):
    axs[1][0].text(x = 0.4,y = i/len(seq)+0.01,s = c, fontsize = 10)
alignment = []
i=0
for x,g in itertools.groupby(kmers):
    alignment += [i]*len(list(g))
    i+=1
alignment = np.asarray(alignment)
axs[1][1].plot(alignment[-1] - alignment,color = "black")
axs[1][1].set_ylim(ymin = -0.5,ymax = alignment[-1]+0.5)
fig.savefig(f"{out_f}/alignment1.png",dpi = 300, bbox_inches = "tight")
# save pdf
fig.savefig(f"{out_f}/alignment1.pdf",dpi = 300, bbox_inches = "tight",format = "pdf")
#save eps
fig.savefig(f"{out_f}/alignment1.eps",dpi = 300, bbox_inches = "tight",format = "eps")

## Print the Marcus suggestion
fig,axs = plt.subplots(nrows = 2, 
                       ncols = 2, 
                       gridspec_kw={'width_ratios': [1, 10],
                                    "height_ratios":[1,5],
                                    'wspace':0, 
                                    'hspace':0},
                       figsize = (6,6))
for i, ax in enumerate(fig.axes):
    ax.set_xticklabels([])
    ax.set_yticklabels([])
axs[0][0].set_axis_off()
axs[0][1].plot(signal)
axs[0][1].xaxis.set_visible(False)
for i,c in enumerate(seq):
    axs[1][0].text(x = 0.4,y = i/len(seq)+0.01,s = c, fontsize = 10)
alignment = []
i=0
for x,g in itertools.groupby(kmers):
    alignment += [i]*len(list(g))
    i+=1
alignment = np.asarray(alignment)
axs[1][1].plot(alignment[-1] - alignment,color = "black")
axs[1][1].set_ylim(ymin = -0.5,ymax = alignment[-1]+0.5)
axs[1][1].fill_between(x = np.arange(len(signal)),
                       y1 = -0.5,
                       y2 = alignment[-1]+0.5,
                       color = "grey")
fig.savefig(f"{out_f}/alignment2.png",dpi = 300, bbox_inches = "tight")
# save pdf
fig.savefig(f"{out_f}/alignment2.pdf",dpi = 300, bbox_inches = "tight",format = "pdf")
#save eps
fig.savefig(f"{out_f}/alignment2.eps",dpi = 300, bbox_inches = "tight",format = "eps")

## Print the Marcus suggestion
fig,axs = plt.subplots(nrows = 2, 
                       ncols = 2, 
                       gridspec_kw={'width_ratios': [1, 10],
                                    "height_ratios":[1,5],
                                    'wspace':0, 
                                    'hspace':0},
                       figsize = (6,6))
for i, ax in enumerate(fig.axes):
    ax.set_xticklabels([])
    ax.set_yticklabels([])
axs[0][0].set_axis_off()
axs[0][1].plot(signal)
axs[0][1].xaxis.set_visible(False)
for i,c in enumerate(seq):
    axs[1][0].text(x = 0.4,y = i/len(seq)+0.01,s = c, fontsize = 10)
alignment = []
i=0
for x,g in itertools.groupby(kmers):
    alignment += [i]*len(list(g))
    i+=1
alignment = np.asarray(alignment)
alignment = alignment[-1] - alignment
axs[1][1].plot(alignment,color = "black")
axs[1][1].fill_between(x = np.arange(len(signal)),
                       y1 = alignment - 3,
                       y2 = alignment + 3,
                       color = "grey")
axs[1][1].set_ylim(ymin = -0.5,ymax = alignment[0]+0.5)
fig.savefig(f"{out_f}/alignment3.png",dpi = 300, bbox_inches = "tight")
# save pdf
fig.savefig(f"{out_f}/alignment3.pdf",dpi = 300, bbox_inches = "tight",format = "pdf")
#save eps
fig.savefig(f"{out_f}/alignment3.eps",dpi = 300, bbox_inches = "tight",format = "eps")



import os
import sys
import h5py
import xron
import toml
import torch
import shutil
import inspect
import argparse
import numpy as np
from typing import Callable
from datetime import datetime
from itertools import groupby 
from functools import partial
from fast_ctc_decode import beam_search
from boostnano.boostnano_model import CSM
from boostnano.boostnano_eval import evaluator
from xron.xron_model import CRNN, CONFIG
from xron.xron_train_base import load_config
from xron.utils.seq_op import raw2seq,fast5_iter,norm_by_noisiest_section,med_normalization,list2string
from xron.utils.easy_assembler import simple_assembly_qs


##Debug module
from timeit import default_timer as timer
##

def load(model_folder,net,device = 'cuda'):
    ckpt_file = os.path.join(model_folder,'checkpoint')
    with open(ckpt_file,'r') as f:
        latest_ckpt = f.readline().strip().split(':')[1]
    ckpt = torch.load(os.path.join(model_folder,latest_ckpt),
                      map_location=device)
    if isinstance(ckpt,dict):
        net.load_state_dict(ckpt["encoder"])
    else:
        net.load_state_dict(ckpt)
    net.to(device)

def chunk_feeder(fast5_f,config,boostnano_evaluator = None):
    iterator = fast5_iter(fast5_f,tqdm_bar = True)
    e = config.EVAL
    batch_size,chunk_len,device,offset,overlay = e['batch_size'],e['chunk_len'],e['device'],e['offset'],e['overlay']
    chunks,meta_info = [],[]
    for read_h,signal,fast5_f,read_id in iterator:
        read_len = len(signal)
        if e['mode'] == 'rna' or e['mode'] == 'rna-meth':
            if boostnano_evaluator is None:
                signal = signal[::-1]
                signal = norm_by_noisiest_section(signal,offset = offset)[0].astype(np.float32)
            else:
                (decoded,path,locs) = boostnano_evaluator.eval_sig(signal,1000)
                locs = read_len - locs
                signal = signal[::-1][:locs[0]]
                if len(signal) == 0:
                    continue
                signal = med_normalization(signal).astype(np.float32)
                read_len = len(signal)
        else:
            signal = norm_by_noisiest_section(signal,offset = offset)[0].astype(np.float32)
        if read_len <= overlay:
            current_chunks = [signal]
        else:
            current_chunks = [signal[i:i+chunk_len] for i in np.arange(0,read_len-overlay,chunk_len-overlay)]
        last_chunk = current_chunks[-1]
        last_chunk_len = len(last_chunk)
        current_chunks[-1]= np.pad(last_chunk,(0,chunk_len-last_chunk_len),'constant',constant_values = (0,0))
        chunks += current_chunks
        meta_info += [(fast5_f,read_id,chunk_len)]*(len(current_chunks)-1) 
        meta_info += [(fast5_f,read_id,last_chunk_len)]
        while len(chunks) >= batch_size:
            curr_chunks = chunks[:batch_size]
            curr_meta = meta_info[:batch_size]
            yield torch.from_numpy(np.stack(curr_chunks,axis = 0)[:,None,:].astype(np.float32)).to(device),curr_meta
            chunks = chunks[batch_size:]
            meta_info = meta_info[batch_size:]
    if len(chunks):
        chunks += [chunks[-1]]*(batch_size - len(chunks))
        meta_info += [(None,None,None)]*(batch_size - len(chunks))
        yield torch.from_numpy(np.stack(chunks,axis = 0)[:,None,:].astype(np.float32)).to(device),meta_info

def qs(consensus, consensus_qs, output_standard='phred+33'):
    """
    Calculate the quality score for the consensus read.

    Args:
        consensus (Int): 2D Matrix (bases, read length) given the count of base on each position.
        consensus_qs (Float): 1D Vector given the mean of the difference between the highest logit and second highest logit.
        output_standard (str, optional): Defaults to 'phred+33'. Quality score output format.

    Returns:
        quality score: Return the queality score as int or string depending on the format.
    """

    sort_ind = np.argsort(consensus, axis=0)
    L = consensus.shape[1]
    sorted_consensus = consensus[sort_ind, np.arange(L)[np.newaxis, :]]
    sorted_consensus_qs = consensus_qs[sort_ind, np.arange(L)[np.newaxis, :]]
    quality_score = 10 * (np.log10((sorted_consensus[-1, :] + 1) / (
        sorted_consensus[-2, :] + 1))) + sorted_consensus_qs[-1, :] / sorted_consensus[-1, :] / np.log(10)
    if output_standard == 'number':
        return quality_score.astype(int)
    elif output_standard == 'phred+33':
        q_string = [chr(x + 33) for x in quality_score.astype(int)]
        return ''.join(q_string)

def viterbi_decode(logits:torch.tensor):
    """
    Viterbi decdoing algorithm

    Parameters
    ----------
    logits : torch.tensor
        Shape L-N-C

    Returns
    -------
        sequence: A length N list contains final decoded sequence.
        moves: A length N list contains the moves array.

    """
    sequence = np.argmax(logits,axis = 2)
    sequence = sequence.T #L,N -> N,L
    sequence,moves = raw2seq(sequence)
    return sequence,list(moves.astype(int))

def consensus_call(seqs,assembly_func,metas,moves,qcs,strides,hiddens = None, logits = None,jump_ratio = None):
    if hiddens:
        curr = [x for x in zip(seqs,metas,qcs,moves,hiddens,logits) if x[1][1] == metas[0][1]]
        curr_seqs,curr_meta,qs_list,moves,hiddens,logits = zip(*curr)
        pad_length = len(moves[0]) - curr_meta[-1][2]//strides
        pad_length_move = len(moves[0]) - np.max([-1] + list(np.where(moves[-1])[0]))-1
        pad_length = min(pad_length,pad_length_move)
        # This pad length gurantee that sum(move) = len(seq), as sometime the base can be called in padding region.
        hiddens = np.asarray(hiddens)
        hiddens = hiddens.reshape((-1,hiddens.shape[-1]))[:-pad_length]
        logits = np.asarray(logits)
        logits = logits.reshape((-1,logits.shape[-1]))[:-pad_length]
    else:
        curr = [x for x in zip(seqs,metas,qcs,moves) if x[1][1] == metas[0][1]]
        curr_seqs,curr_meta,qs_list,moves = zip(*curr)
        pad_length = len(moves[0]) - curr_meta[-1][2]//strides
    css = assembly_func(curr_seqs,qs_list = qs_list,jump_step_ratio = jump_ratio)
    curr_move = np.asarray(moves).flatten()
    fast5f = set([x[0] for x in curr_meta])
    id = set([x[1] for x in curr_meta])
    assert len(fast5f)==1 and len(id)==1
    return css,id.pop(),fast5f.pop(),curr_move[:-pad_length],hiddens,logits

def all_equal(iterable):
    g = groupby(iterable)
    return next(g, True) and not next(g, False)

class Writer(object):
    def __init__(self,write_dest,config):
        self.css_seqs = []
        self.read_ids = []
        self.fast5_fs = []
        self.qs = []
        self.moves = []
        self.hiddens = []
        self.logits = []
        self.count = 0
        self.config = config
        self.base_dict = {i:b for i,b, in enumerate(config.CTC['alphabeta'])}
        self.prefix = write_dest
        self.prefix_fastq = os.path.join(self.prefix,'fastqs')
        os.makedirs(self.prefix_fastq,exist_ok = True)
        self.prefix_fast5 = os.path.join(self.prefix,'fast5s')
        self.seq_batch = config.EVAL['seq_batch']
        self.format = config.EVAL['format']
        if self.format == "fast5":
            os.makedirs(self.prefix_fast5,exist_ok = True)
        self.stride = config.CNN['Layers'][-1]['stride']
        self.chunk_length = config.EVAL['chunk_len']
        
    def add(self,css,read_id,fast5f,move,hidden = None,logits = None):
        if(len(css)) == 2:
            self.css_seqs.append(list2string(np.argmax(css[0],axis=0),base_type = self.base_dict))
            self.qs.append(qs(css[0],css[1]))
        else:
            self.css_seqs.append(list2string(np.argmax(css,axis = 0),base_type = self.base_dict))
        self.moves.append(move)
        self.read_ids.append(read_id)
        self.fast5_fs.append(fast5f)
        if hidden is not None:
            self.hiddens.append(hidden)
        if logits is not None:
            self.logits.append(logits)
        if len(self.css_seqs) >= self.seq_batch:
            self.flush()
    
    def flush(self):
        if not len(self.css_seqs):
            #every thing has been flushed out.
            return 
        if self.format == "fastq":
            self.write_fastq("%s/%d.fastq"%(self.prefix_fastq,self.count))
        if self.format == "fast5":
            self.write_fastq("%s/%d.fastq"%(self.prefix_fastq,self.count))
            self.write_fast5(self.prefix_fast5)
        self.count += 1
        self.css_seqs = []
        self.read_ids = []
        self.fast5_fs = []
        self.moves = []
        self.hiddens = []
        self.qs = []
        self.logits = []
        
    def write_fasta(self,output_f):
        with open(output_f,mode = 'w+') as f:
            for seq,name in zip(self.css_seqs,self.read_ids):
                f.write(">%s\n%s\n"%(name,seq))
    
    def write_fast5(self,output_f):
        last_dest = None
        root = None
        for seq,name,fast5,q,move,hidden,logits in zip(self.css_seqs,self.read_ids,self.fast5_fs,self.qs,self.moves,self.hiddens,self.logits):
            dest = os.path.join(output_f,os.path.basename(fast5))
            if not os.path.isfile(dest):
                shutil.copy(fast5,dest)
            if dest != last_dest:
                if root:
                    root.close()
                root = h5py.File(dest,mode='a')
            if 'read_%s'%(name) not in root:
                print("Warning read %s is not found in the fast5 %s"%(name,dest))
                print("An new entry will be created but original sequencing information will be lost.")
                root.create_group("read_%s"%(name))
            if 'read_%s/Analyses'%(name) in root:
                read_h = root['read_%s/Analyses'%(name)]
                existed_basecall = [x for x in list(read_h.keys()) if 'Basecall' in x]
            else:
                existed_basecall = []
                root['read_%s'%(name)].create_group('Analyses')
                read_h = root['read_%s/Analyses'%(name)]
            curr_group = read_h.create_group('Basecall_1D_%03d'%(len(existed_basecall)))
            curr_group.attrs.create('name','Xron')
            curr_group.attrs.create('time_stamp','%s'%(datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")))
            result_h = curr_group.create_group('BaseCalled_template')
            result_h.create_dataset("Fastq",data = "@%s\n%s\n+\n%s\n"%(name,seq,q))
            result_h.create_dataset("Move",data = move,dtype='i8')
            result_h.create_dataset("Hidden",data = hidden, dtype = "f")
            result_h.create_dataset("Logits",data = logits, dtype = "f")
            summary_h = curr_group.create_group("Summary")
            summary_h = summary_h.create_group("basecall_1d_template")
            summary_h.attrs.create('block_stride',self.stride)
            summary_h.attrs.create('chunk_length',self.chunk_length)
            last_dest = dest
        if root:
            root.close()
    
    def write_fastq(self,output_f):
        with open(output_f,mode = 'w+') as f:
            for seq,name,q in zip(self.css_seqs,self.read_ids,self.qs):
                f.write("@%s\n%s\n+\n%s\n"%(name,seq,q))
    
    def write_config(self,output_f):
        out_config = {"CTC":self.config.CTC,"EVAL":self.config.EVAL}
        config_f = os.path.join(output_f,'config.toml')
        with open(config_f,mode = 'w+') as f:
            toml.dump(out_config,f)

class Evaluator(object):
    def __init__(self,
                 net:CRNN,
                 config:CONFIG,
                 assembly_function:Callable,
                 writer:Writer):
        """

        Parameters
        ----------
        nets : CRNN
            A CRNN network instance.
        config: CONFIG
            A CONFIG class contains basecall configurations.
        assembly_function: Callable
            The assembly function used to assemble chunks into sequence
        writer: Writer:
            The writer function to write the output.
        Returns
        -------
        None.

        """
        self.net = net
        self.config = config
        self.result_collections = {'seqs':[],'metas':[],'moves':[],'qcs':[]}
        if config.EVAL['format'] == 'fast5':
            self.activation = {}
            def get_activation(name):
                def hook(model,input,output):
                    self.activation[name] = output.permute([1,0,2]).cpu().detach().numpy()
                return hook
            net.net[-4].register_forward_hook(get_activation('fnn1_out_linear'))
            self.result_collections['hiddens'] = []
            self.result_collections['logits'] = []
        self.elapse_time = {}
        self.assembly_time, self.nn_time, self.writing_time = 0.,0.,0.
        self.writer = writer
        self.assembly_func = assembly_function
    
    @property
    def seqs(self):
        return self.result_collections['seqs']
    
    @property
    def metas(self):
        return self.result_collections['metas']
    
    @property
    def moves(self):
        return self.result_collections['moves']
    
    @property
    #The quality scores: biggest logits
    def qcs(self):
        return self.result_collections['qcs']
    
    @property
    def logits(self):
        if 'logits' in self.result_collections.keys():
            return self.result_collections['logits']
        else:
            return None
    
    @property
    def hiddens(self):
        if 'hiddens' in self.result_collections.keys():
            return self.result_collections['hiddens']
        else:
            return None
        
    def _timeit(func):
        def wrap(self,*args,**kwargs):
            start = timer()
            func(self,*args,**kwargs)
            if func not in self.elapse_time.keys():
                self.elapse_time[func] = timer() - start
            else:
                self.elapse_time[func] += timer() - start
        return wrap
        
    @_timeit
    def run_nn(self,batch):
        start = timer()
        self.curr_logits = self.net(batch).cpu().numpy() #TODO send the logits into a queue to enable multi-threading
        self.nn_time += timer() - start
    
    def beam_decode(self,posteriors):
        L,N,C = posteriors.shape
        move = [np.zeros(L,dtype = int) for _ in np.arange(N)]
        seqs = []
        for i in np.arange(N):
            p = np.exp(posteriors[:,i,:])
            seq,path = beam_search(p,
                                   "N"+self.config.CTC["alphabeta"],
                                   beam_size = self.config.CTC["beam"],
                                   beam_cut_threshold = self.config.CTC["beam_cut_threshold"])
            move[i][path] = 1
            seqs.append([self.config.CTC["alphabeta"].index(x) for x in seq])
        return seqs,move
    
    def run_once(self,batch,meta):
        self.run_nn(batch)
        start = timer()
        if self.config.CTC["beam"] <= 1:
            sequence,move = viterbi_decode(self.curr_logits)
        else:
            sequence,move = self.beam_decode(self.curr_logits)
        self.result_collections['qcs'] += list(np.max(self.curr_logits,axis = 2).T)
        self.result_collections['seqs'] += sequence
        self.result_collections['metas'] += meta
        self.result_collections['moves'] += move
        if self.config.EVAL['format'] == 'fast5':
            self.result_collections['hiddens'] += list(self.activation['fnn1_out_linear'])
            self.result_collections['logits'] += list(np.transpose(self.curr_logits,(1,0,2)))
        self.assembly_time += timer() - start
        self._call()
    
    def _call(self):
        if (None,None,None) in self.metas:
            print("Remove the padding chunks.")
            self._update_remain()
        while not all_equal([x[1] for x in self.metas]):
            self.call_once()
    @_timeit
    def call_once(self):
        start = timer()
        if len(self.seqs) == 0:
            return
        r = consensus_call(self.seqs,
                           self.assembly_func,
                           self.metas,
                           self.moves,
                           qcs = self.qcs,
                           strides = self.config.CNN['Layers'][-1]['stride'], 
                           hiddens = self.hiddens, 
                           logits = self.logits,
                           jump_ratio= 1-float(self.config.EVAL['overlay'])/self.config.EVAL['chunk_len'])
        self.assembly_time += timer() - start
        start = timer()
        self.writer.add(*r)
        writing_duration = timer()-start
        self.writing_time += writing_duration
        self._update_remain(key = self.metas[0][1])
    
    @_timeit
    def _update_remain(self,key = None):
        start = timer()
        remain = [x for x in zip(*list(self.result_collections.values())) if x[1][1] != key] #Update based on metas name (x[1] = metas)
        remain = list(zip(*remain))
        if len(remain) == 0:
            return
        for i,k in enumerate(list(self.result_collections.keys())):
            self.result_collections[k] = list(remain[i])
        self.assembly_time += timer() - start

def locate_model(model_f):
    if os.path.exists(model_f):
        return model_f
    package_f = xron.__path__[0]
    print(package_f)
    if os.path.exists(os.path.join(package_f,model_f)):
        return os.path.join(package_f,model_f)
    if os.path.exists(os.path.join(package_f,'models',model_f)):
        return os.path.join(package_f,'models',model_f)
    else:
        raise ValueError(f"Cannot find the model file at {model_f}.")

def main(args):
    if args.threads:
        torch.set_num_threads(args.threads)
    class CTC_CONFIG(CONFIG):
        CTC = {"beam_cut_threshold":0.05,
               "alphabeta": "ACGTM",
               "mode":"rna",
               "beam":args.beam}
    class CALL_CONFIG(CTC_CONFIG):
        EVAL = {"batch_size":args.batch_size,
                "chunk_len":args.chunk_len,
                'device':args.device,
                'assembly_method':args.assembly_method,
                'seq_batch':4000,
                'mode':'rna',
                'format':'fast5' if args.fast5 else 'fastq',
                'offset':args.offset,
                'diff_norm':args.diff_norm}        #Diffnorm is deprecated setting this to True will have no effect
    config = CALL_CONFIG()
    print("Construct and load the model.")
    model_f =locate_model(args.model_folder)
    config_old = load_config(os.path.join(model_f,"config.toml"))
    config_old.EVAL = config.EVAL #Overwrite training config.
    config_old.CTC = config.CTC
    config = config_old
    if args.config:
        config = load_config(args.config)
    net = CRNN(config)
    load(model_f,net,device = args.device)
    print("Begin basecall.")
    net.eval()
    boostnano_evaluator = None
    if config.EVAL['mode'] == 'rna' or config.EVAL['mode'] == 'rna-meth':
        if args.boostnano:
            print("Loading BoostNano model.")
            project_f = os.path.dirname(os.path.dirname(inspect.getfile(CSM)))
            model_f = os.path.join(project_f,'boostnano','model')
            boostnano_net = CSM()
            boostnano_evaluator = evaluator(boostnano_net,model_f)
    config.EVAL['overlay'] = int(args.overlay_ratio * args.chunk_len)
    df = chunk_feeder(args.input, config, boostnano_evaluator)
    writer = Writer(args.output,config)
    writer.write_config(args.output)
    assembly_func = partial(simple_assembly_qs,
                            kernal = args.assembly_method,
                            alphabeta = config.CTC["alphabeta"])
    caller = Evaluator(net,config,assembly_function = assembly_func,writer = writer)
    with torch.no_grad():
        for chunk,meta in df:
            caller.run_once(chunk,meta)
    start_w = timer()
    caller.call_once()
    writer.flush()
    caller.writing_time += timer()-start_w
    print("NN_time:%f,assembly_time:%f,writing_time:%f"%(caller.nn_time,caller.assembly_time,caller.writing_time))

def add_arguments(parser):
    parser.add_argument('-i', '--input', required = True,
                        help = "The input folder contains the fast5 files.")
    parser.add_argument('-m', '--model_folder', required = True,
                        help = "The folder contains the model, can also be model identifier including ENEYFT, ENE, ENEHEK. Recommend to use ENEYFT model.")
    parser.add_argument('-o', '--output', required = True,
                        help = "The output folder.")
    parser.add_argument('--fast5',action = "store_true",dest = "fast5",
                        help = "If output fast5 files.")
    parser.add_argument('--device', default = None,
                        help="The device used for training, can be cpu or cuda.")
    parser.add_argument('--batch_size', default = None, type = int,
                        help="Evaluation batch size, default use the maximum size of the GPU memory, batch size and memory consume relationship: 200:6.4GB, 400:11GB, 800:20GB, 1200:30GB.")
    parser.add_argument('--chunk_len', default = 2000, type = int,
                        help="The length of each chunk signal.")
    parser.add_argument('--overlay_ratio', default = 0.0, type = float,
                        help="The ratio of the overlay between two chunks.")
    parser.add_argument('--config', default = None,
                        help = "Training configuration.")
    parser.add_argument('--threads', type = int, default = None,
                        help = "Number of threads used by Pytorch")
    parser.add_argument('--assembly_method',type = str, default = "stick",
                        help = "Assembly method used, can be glue, global, simple and stick.")
    parser.add_argument('--offset',type = float,default = 0.0,
                        help = "Manual set a offset to the normalized signal.")
    parser.add_argument('--diff_norm', action="store_true", dest="diff_norm",
                        help = "Turn on the differential normalization.")
    parser.add_argument('--beam', type = int, default = 1,
                        help = "Beam size of the beam search decoder, default\
                        is 1 where Viterbi decoder is used.")
    parser.add_argument('--boostnano', action="store_true", dest="boostnano",
                        help = "Enable boostnano preprocessing.")

def post_args(args):
    MEMORY_PER_BATCH_PER_SIGNAL=15000. #B
    DEFAULT_BATCH_SIZE = 1200
    if args.device is None:
        if torch.cuda.is_available():
            args.device = "cuda:0"
            t = torch.cuda.get_device_properties(0).total_memory
            if args.batch_size is None:
                args.batch_size = int(t/MEMORY_PER_BATCH_PER_SIGNAL/args.chunk_len//100*100)
                print("Auto configure to use %d batch_size for a total of %.1f GB memory."%(args.batch_size,t/1024**3))
            
        elif torch.backends.mps.is_available():
            #this is a apple m GPU
            args.device = "mps"
            if args.batch_size is None:
                args.batch_size = 600
                print("MPS backend detected, the batch_size is setting to default %d"%(args.batch_size))
        else:
            args.device = "cpu"
            if args.batch_size is None:
                args.batch_size = DEFAULT_BATCH_SIZE
                print("No GPU is detected, the batch_size is setting to default %d"%(args.batch_size))
    os.makedirs(args.output,exist_ok=True)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Call a xron model on dataset.')
    add_arguments(parser)
    args = parser.parse_args(sys.argv[1:])
    post_args(args)
    main(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 07:34:49 2022

@author: heavens
"""
from matplotlib import pyplot as plt
import numpy as np

def auc_plot(TP,FP,axs,add_head = False,print_AUC = True,**kwargs):
    c = kwargs['color'] if 'color' in kwargs.keys() else None
    label = kwargs['label'] if 'color' in kwargs.keys() else None
    linewidth = kwargs['linewidth'] if 'linewidth' in kwargs.keys() else 1
    TP = np.asarray(TP)
    FP = np.asarray(FP)
    assert np.all(TP[1:] - TP[:-1]>=0)
    assert np.all(FP[1:] - FP[:-1]>=0)
    if add_head:
        TP = np.concatenate(([0],TP,[1]))
        FP = np.concatenate(([0],FP,[1]))
    axs.plot(FP[:2],TP[:2],color = c,label = label,linewidth = linewidth)
    axs.set_xlabel("False Positive")
    axs.set_ylabel("True Positive")
    axs.set_xlim([-0.05, 1.05])
    axs.set_ylim([-0.05, 1.05])
    if print_AUC:
        axs.text(x = 0.8,y = 0,s = "AUC = %.2f"%(AUC(TP,FP)))
    axs.plot([-0.05,1.05],[-0.05,1.05],color = 'grey')

def AUC(TP,FP):
    """Calculate Area under curve given the true positive and false positive
    array
    """
    TP = TP[::-1] if TP[0]>TP[-1] else TP
    FP = FP[::-1] if FP[0]>FP[-1] else FP
    TP = [0] + TP if TP[0] != 0 else TP
    TP = TP + [1] if TP[-1] != 1 else TP
    FP = [0] + FP if FP[0] != 0 else FP
    FP = FP + [1] if FP[-1] != 1 else FP
    FP = np.asarray(FP)
    TP = np.asarray(TP)
    return np.sum((TP[1:] + TP[:-1])*(FP[1:]-FP[:-1])/2)
    


# Copyright 2017 The Xron Authors. All Rights Reserved.
#
#This Source Code Form is subject to the terms of the Mozilla Public
#License, v. 2.0. If a copy of the MPL was not distributed with this
#file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
#Created on Mon Aug 14 18:38:18 2017
import argparse
import sys
import logging
from os import path
import xron
from xron import xron_eval
from xron import xron_train_supervised
from xron import xron_init
from xron.utils import prepare_chunk
from xron.nrhmm import hmm_relabel

def check_init(init_function):
    def decorator(function):
        def wrapper(*args, **kwargs):
            init_function()
            result = function(*args, **kwargs)
            return result
        return wrapper
    return decorator

def check_model():
    model_path = path.join(xron.__path__[0], 'models')
    if not path.exists(model_path):
        logging.error('Models not found. Please run "xron init" first.')
        sys.exit(1)

@check_init(check_model)
def evaluation(args):
    xron_eval.post_args(args)
    xron_eval.main(args)

@check_init(check_model)
def export(args):
    prepare_chunk.post_args(args)
    prepare_chunk.extract(args)

@check_init(check_model)
def train(args):
    xron_train_supervised.post_args(args)
    xron_train_supervised.main(args)

@check_init(check_model)
def relabel(args):
    hmm_relabel.post_args(args)
    hmm_relabel.main(args)

def main(arguments=sys.argv[1:]):
    parser = argparse.ArgumentParser(prog='xron', description='A deep neural network basecaller that achieve methylation.')
    parser.add_argument('-v','--version',action='version',version='Xron version '+xron.__version__,help="Print out the version.")
    subparsers = parser.add_subparsers(title='sub command', help='sub command help')

    # parser for 'init' command
    parser_init = subparsers.add_parser('init', description='Initialize xron package, need to run this when first time runnning xron',
                                        help='Initialize xron package, need to run this when first time runnning xron')
    parser_init.set_defaults(func=xron_init.get_models)

    # parser for 'call' command
    parser_call = subparsers.add_parser('call', description='Perform basecalling', help='Perform basecalling.')
    xron_eval.add_arguments(parser_call)
    parser_call.set_defaults(func=evaluation)
    
    # parser for 'extract' command
    parser_export = subparsers.add_parser('prepare', description='Prepare the training dataset by aligning it to the reference genome, it is an equivalent command to resquiggle when --extract_seq flag is set.',
                                          help='Realign the sequence to the reference genome and extract the signal chunk and label for training.')
    prepare_chunk.add_arguments(parser_export)
    parser_export.set_defaults(func=export)

    # parser for 'relabel' command
    parser_relabel = subparsers.add_parser('relabel', description='Relabel the training dataset using the pretraiend NHMM model.',
                                             help='Relabel the training dataset using the pretraiend NHMM model.')
    hmm_relabel.add_arguments(parser_relabel)
    parser_relabel.set_defaults(func=relabel)
    
    # parser for 'train' command
    parser_train = subparsers.add_parser('train', description='Training a model in several ways: embedding, supervised and reinforce', help='Train a model.')
    xron_train_supervised.add_arguments(parser_train)
    parser_train.set_defaults(func=train)

    args = parser.parse_args(arguments)
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    #print(sys.argv[1:])
    main()


"""
Created on Wed Apr 27 09:01:02 2022
This script is writen to load a HMM model and used it to summirize the output 
of a basecalled dataset.
@author: Haotian Teng
"""
import os
import time
import toml
import torch
import itertools
import numpy as np
from scipy.stats import binom
from tqdm import tqdm
from matplotlib import pyplot as plt
from xron.nrhmm.hmm import GaussianEmissions, RHMM
from xron.nrhmm.kmer2seq import fixing_looping_path
from xron.nrhmm.hmm_input import Kmer2Transition, Kmer_Dataset, Normalizer
from xron.utils.plot_op import auc_plot
from torchvision import transforms
from torch.utils.data.dataloader import DataLoader
from xron.xron_train_base import DeviceDataLoader
from xron.utils.seq_op import Methylation_DP_Aligner
### Test module
from time import time

def kmers2seq(kmers,idx2kmer):
    merged = [g for g,_ in itertools.groupby(kmers)]
    seqs = [idx2kmer[x][0] for x in merged]
    return ''.join(seqs) + idx2kmer[merged[-1]][1:]

def print_pore_model(emission_means,idx2kmer,f = None):
    for i,mean in enumerate(emission_means):
        if f:
            f.write("%s: %.2f\n"%(idx2kmer[i],mean))
        else:
            print("%s: %.2f"%(idx2kmer[i],mean))
            
def load_pore_model(pore_model_f):
    pore_model = dict()
    with open(pore_model_f) as f:
        for line in f:
            split_line = line.split(":")
            pore_model[split_line[0]] = float(split_line[1])
    return pore_model

def multiple_testing(TP,FP,n = 100):
    TP_n = [1-binom.cdf(int(n/2),n,x) for x in TP]
    FP_n = [1-binom.cdf(int(n/2),n,x) for x in FP]
    return TP_n,FP_n

def posterior_decode(posterior,
                     idx2kmer,
                     kmer2idx,
                     AM_threshold:float = 1.):
    B,T,N = posterior.shape
    AM_ratio = []
    for p in posterior:
        path = torch.argmax(p,dim = 1)
        kmers = [idx2kmer[x.item()] for x in path]
        A_kmers,M_kmers = [],[]
        idxs = []
        for i,kmer in enumerate(kmers):
            if kmer[2] == 'A' or kmer[2] == 'M':
                kmer = ''.join([x if i!=2 else 'A' for i,x in enumerate(list(kmer)) ])
                A_kmers.append(kmer)
                kmer = ''.join([x if i!=2 else 'M' for i,x in enumerate(list(kmer)) ])
                M_kmers.append(kmer)
                idxs.append(i)
        A_kmers = torch.tensor([kmer2idx[x] for x in A_kmers])
        M_kmers = torch.tensor([kmer2idx[x] for x in M_kmers])
        idxs = torch.tensor(idxs)
        A_p = p[idxs,A_kmers]
        M_p = p[idxs,M_kmers]
        cut = torch.log(A_p/M_p)<AM_threshold
        AM_ratio.append(torch.sum(cut).item()/(len(cut)))
    return AM_ratio
        

def compare_pore_models(pmf1:str, pmf2:str, selection:str):
    """
    Compare two pore models

    Parameters
    ----------
    pmf1 : str
        The string contains the file of the pore model.
    pmf2 : str
        The string contains the file of the second pore model.
    selection : str
        The magic string to filter the kmers to compare with, e.g. M!A, means
        most contain M and must not contain A in the kmer.

    Returns
    -------
    the difference of the kmers.
    """
    pm1 = load_pore_model(pmf1)
    pm2 = load_pore_model(pmf2)
    demand_char = [x for x in selection.split("!")[0]]
    filter_char = [x for x in selection.split("!")[1]]
    selected_kmers = [(val,pm2[key]) for key,val in pm1.items() if all([d in key for d in demand_char] + [f not in key for f in filter_char]) ]
    k1,k2 = list(zip(*selected_kmers))
    return np.asarray(k1),np.asarray(k2)

def run_test(args):
    model = torch.load(args.ckpt_f,
                       map_location = torch.device(args.device))
    # emission = GaussianEmissions(model['hmm']['emission.means'].cpu().numpy(), model['hmm']['emission.certainty'].cpu().numpy())
    emission = GaussianEmissions(model['hmm']['emission.means'].cpu().numpy(),model['hmm']['emission.cov'].cpu().numpy())
    print(model['hmm']['emission.cov'].cpu().numpy())
    hmm = RHMM(emission,normalize_transition=False,device = args.device)
    # sns.distplot(hmm.emission.certainty.cpu().numpy())
    # print(model['hmm']['emission.certainty'].cpu().numpy())
    config = toml.load(os.path.join(args.input,"config.toml"))
    print("Readout the pore model.")
    with open(args.ckpt_f+"_poremodel", "w+") as f:
        print_pore_model(model['hmm']['emission.means'],config['idx2kmer'],f)
    chunks = np.load(os.path.join(args.input,"chunks.npy"),mmap_mode="r")
    n_samples, sig_len = chunks.shape
    durations = np.load(os.path.join(args.input,"durations.npy"),mmap_mode = "r")
    idx2kmer = config['idx2kmer']
    kmer2idx = config['kmer2idx_dict']
    signal_collection = [[] for _ in np.arange(emission.N)]
    m6a_count = [0,0]
    thresholds = np.asarray(np.arange(-300,300,1)).astype(np.float)
    rocs = {x:[] for x in thresholds}
    mse,mse_renorm,m6a_ratios = [],[],[]
    kmers = np.load(os.path.join(args.input,"kmers.npy"))
    aligner = Methylation_DP_Aligner(base_alternation = {'M':'A'})
    k2t = Kmer2Transition(alphabeta = config['alphabeta'],
                          k = config['k'],
                          T_max = config['chunk_len'],
                          kmer2idx = config['kmer2idx_dict'],
                          idx2kmer = config['idx2kmer'],
                          neighbour_kmer = args.neighbour ,
                          base_alternation = {"M":"A"}, 
                          base_prior = {x:args.m_prior for x in config['alphabeta']},
                          kmer_replacement = True)
    dataset = Kmer_Dataset(chunks, durations, kmers,transform=transforms.Compose([k2t]))
    loader = DataLoader(dataset,batch_size = args.batch_size, shuffle = False)
    loader = DeviceDataLoader(loader,device = args.device)
    norm = Normalizer(use_dwell = True,no_scale = True)
    count = 0
    for i_batch, batch in tqdm(enumerate(loader)):
        signal_batch = batch['signal']
        duration_batch = batch['duration']
        transition_batch = batch['labels']
        kmers_batch = batch['kmers']
        renorm_batch = signal_batch
        for j in tqdm(np.arange(args.n_renorm+1),desc = "Renorm the signal:"):
            with torch.no_grad():
                path,logit = hmm.viterbi_decode(renorm_batch, duration_batch, transition_batch)
            disagreement = torch.sum(path != kmers_batch,axis = 1)/duration_batch
            print("The disagreement between the doecded path and basecaller path: %.2f"%(torch.mean(disagreement)))
            rc_signal = np.asarray([[hmm.emission.means[x].item() for x in p] for p in path.cpu().numpy()])
            rc_signal_kmer = np.asarray([[hmm.emission.means[x].item() for x in p] for p in kmers_batch.cpu().numpy()])
            renorm_batch = norm(renorm_batch.cpu().numpy()[:,:,0],rc_signal, duration_batch.cpu().numpy(),path.cpu().numpy())
            renorm_batch = torch.from_numpy(renorm_batch).unsqueeze(dim = -1).to(args.device)
        renorm_batch = renorm_batch.squeeze(dim = -1).cpu().numpy()
        with torch.no_grad():
            gamma = hmm.expectation(torch.tensor(renorm_batch).float().unsqueeze(dim = -1).to(args.device), duration_batch, transition_batch)
            for t in tqdm(thresholds):
                rocs[t]+= posterior_decode(gamma,idx2kmer,kmer2idx,t)
        for i in np.arange(args.batch_size):
            if args.visual:
                fig,axs = plt.subplots()
                orig_sig = signal_batch[i][:duration_batch[i]].detach().cpu().numpy()
                rescale_sig = renorm_batch[i]
                axs.plot(orig_sig,c = 'g',label = "original signal")
                axs.plot(rescale_sig,c = 'r', label = "Rescaled signal")
                # rc_signal_orig = [hmm.emission.means[x].item() for x in kmers_batch[i]
                axs.plot(rc_signal[i][:duration_batch[i]],color = 'blue',label = 'reconstructed hmm',alpha = 0.7)
                axs.plot(rc_signal_kmer[i][:duration_batch[i]],color = 'yellow',label = 'reconstructed basecaller',alpha = 0.7)
                # axs.plot(rc_signal_orig[:duration_batch[i]],color = "red",label = "original label")
                diff = orig_sig[:,0] - rc_signal[i][:duration_batch[i]]
                diff_renorm = rescale_sig[:duration_batch[i]] - rc_signal[i][:duration_batch[i]]
                mse.append(np.mean(diff ** 2))
                mse_renorm.append(np.mean(diff_renorm ** 2))
                plt.legend()
                plt.xlabel("Time")
                plt.ylabel("Signal")
                plt.figure()
                dwell_length = np.asarray([len(list(x)) for _,x in itertools.groupby(path[i][:duration_batch[i]])])
                sns.histplot(dwell_length[dwell_length<10])
            deco_seq = kmers2seq(path.cpu().numpy()[i][:duration_batch[i]],idx2kmer)
            orig_seq = kmers2seq(kmers_batch.cpu().numpy()[i][:duration_batch[i]],idx2kmer)
            deco_aln,orig_aln = aligner.align(deco_seq,orig_seq)
            final_seq = aligner.merge(deco_aln,orig_aln)
            m6a_count[0] += final_seq.count('M')
            m6a_count[1] += final_seq.count('A')
            ratio = final_seq.count('M')/(final_seq.count('M')+final_seq.count('A'))
            m6a_ratios.append(ratio)
            print("m6A ratio %.2f"%(ratio))
            if args.visual:
                print("%d sample:"%(i))
                print("Decoded sequence:  %s"%(deco_seq))
                print("Align-corrected sequence: %s"%(final_seq))
                print("Original sequence: %s"%(orig_seq))
                print("MSE error: %.2f, MSE error renorm: %.2f"%(np.mean(mse),np.mean(mse_renorm)))
        count += 1
        if count >= args.n_cases:
            break
    m6a_ratios = np.asarray(m6a_ratios)
    print("m6A ratio %.2f"%(np.mean(m6a_ratios)))
    if args.visual:
        plt.plot(logit[0,np.arange(4000),kmers_batch[0]].detach().cpu().numpy(),label = "Oirignal.")
        plt.plot(logit[0,np.arange(4000),path[0]].detach().cpu().numpy(),label = "Realigned.")
        plt.ylabel("Log probability.")
        plt.xlabel("Time")
    return m6a_ratios,rocs

class Tester(object):
    def __init__(self,model_config):
        self.config = model_config
        self.datasets = {}
        self.loader = None
        self._load_model()
        
    def _load_model(self):
        model = torch.load(self.config.ckpt_f,
                           map_location = torch.device(self.config.device))
        emission = GaussianEmissions(model['hmm']['emission.means'].cpu().numpy(),model['hmm']['emission.cov'].cpu().numpy())
        self.hmm = RHMM(emission,normalize_transition=False,device = self.config.device)
    
    def _export_pore_model(self):
        with open(self.config.ckpt_f+"_poremodel", "w+") as f:
            print_pore_model(self.emission.means,self.data_config['idx2kmer'],f)
    
    def load_data_config(self,confiugre_file):
        config = toml.load(confiugre_file)
        self.data_config = config
        return config
    
    def add_dataset(self,dataset_folder,name = None):
        config = self.load_data_config(os.path.join(dataset_folder,"config.toml"))
        chunks = np.load(os.path.join(dataset_folder,"chunks.npy"),mmap_mode="r")
        n_samples, sig_len = chunks.shape
        durations = np.load(os.path.join(dataset_folder,"durations.npy"),mmap_mode = "r")
        kmers = np.load(os.path.join(dataset_folder,"kmers.npy"),mmap_mode = "r")
        k2t = Kmer2Transition(alphabeta = self.data_config['alphabeta'],
                          k = config['k'],
                          T_max = config['chunk_len'],
                          kmer2idx = config['kmer2idx_dict'],
                          idx2kmer = config['idx2kmer'],
                          neighbour_kmer = self.config.neighbour,
                          base_alternation = {"M":"A"}, 
                          base_prior = {"M":self.config.m_prior},
                          kmer_replacement = True)
        dataset = Kmer_Dataset(chunks, durations, kmers,transform=transforms.Compose([k2t]))
        name = name if name is not None else "dataset %d"%(len(self.datasets.keys())+1)
        self.datasets[name] = dataset

    def load_dataset(self,name,shuffle = True):
        if len(self.datasets.keys()) == 0:
            raise ValueError("No avaliable dataset is found, please add dataset first.")
        dataset = self.datasets[name]
        loader = DataLoader(dataset,batch_size = self.config.batch_size, shuffle = shuffle)
        self.loader = DeviceDataLoader(loader,device = self.config.device)
        return loader
    
    def renormalization(self, 
                        batch,
                        n_renorm = None):
        if n_renorm is None:
            n_renorm = self.config.n_renorm
        norm = Normalizer(use_dwell = self.config.dwell_norm,
                          no_scale = self.config.no_scale)
        signal_batch = batch['signal']
        duration_batch = batch['duration']
        transition_batch = batch['labels']
        kmers_batch = batch['kmers']
        renorm_batch = signal_batch
        for j in tqdm(np.arange(n_renorm),desc = "Renorm the signal:"):
            with torch.no_grad():
                path,logit = self.hmm.viterbi_decode(renorm_batch, duration_batch, transition_batch)
                for i in np.arange(len(path)):
                    seq = kmers2seq(kmers_batch[i],idx2kmer = self.data_config['idx2kmer'])
                    if config.fix_alignment:
                        p = path[i].cpu().numpy()
                        d = duration_batch[i].cpu().numpy()
                        fix_path = fixing_looping_path(p[:d],seq,idx2kmer = self.data_config['idx2kmer'],canonical_base = 'A',modified_base = 'M')
                        path[i,:duration_batch[i]] = torch.from_numpy(fix_path)
            rc_signal = np.asarray([[self.hmm.emission.means[x].item() for x in p] for p in path.cpu().numpy()])
            rc_signal_kmer = np.asarray([[self.hmm.emission.means[x].item() for x in p] for p in kmers_batch.cpu().numpy()])
            renorm_batch = norm(renorm_batch.cpu().numpy()[:,:,0],rc_signal, duration_batch.cpu().numpy(),path.cpu().numpy())
            renorm_batch = torch.from_numpy(renorm_batch).unsqueeze(dim = -1).to(self.config.device)
        return renorm_batch,rc_signal,rc_signal_kmer
    
    def renormalization_test(self,
                             n_batch = 1,
                             plot = False):
        if self.loader is None:
            raise ValueError("No dataset has been loaded, call Tester.load_dataset first.")
        collection = {"RC_MSE":[],"RC_RENORM_MSE":[]}
        for i_batch, batch in tqdm(enumerate(self.loader)):
            if i_batch >= n_batch:
                break
            signal_batch = batch['signal']
            duration_batch = batch['duration']
            transition_batch = batch['labels']
            kmers_batch = batch['kmers']
            renorm_batch,rc_signal,rc_signal_kmer = self.renormalization(batch)
            with torch.no_grad():
                path,logit = self.hmm.viterbi_decode(renorm_batch, duration_batch, transition_batch)
            renorm_batch = renorm_batch.squeeze(dim = -1).cpu().numpy()
            for i in np.arange(self.config.batch_size):
                orig_sig = signal_batch[i][:duration_batch[i]].detach().cpu().numpy()
                rescale_sig = renorm_batch[i]
                diff = orig_sig[:,0] - rc_signal[i][:duration_batch[i]]
                diff_renorm = rescale_sig[:duration_batch[i]] - rc_signal[i][:duration_batch[i]]
                collection['RC_MSE'].append(np.mean(diff ** 2))
                collection['RC_RENORM_MSE'].append(np.mean(diff_renorm ** 2))
                if plot:
                    print("Sequence %d"%(i))
                    if self.config.fix_alignment:
                        seq = kmers2seq(kmers_batch[i],idx2kmer = self.data_config['idx2kmer'])
                        fix_path = fixing_looping_path(path[i,:duration_batch[i]],seq,idx2kmer = self.data_config['idx2kmer'],canonical_base = 'A',modified_base = 'M')
                        print("pred:%s"%(kmers2seq(fix_path,self.data_config['idx2kmer'])))
                    else:
                        print("pred:%s"%(kmers2seq(path[i],self.data_config['idx2kmer'])))
                    print("orig:%s"%(kmers2seq(kmers_batch[i],self.data_config['idx2kmer'])))
                    fig,axs = plt.subplots()
                    axs.plot(orig_sig,c = 'g',label = "original signal")
                    axs.plot(rescale_sig,c = 'r', label = "Rescaled signal")
                    # rc_signal_orig = [hmm.emission.means[x].item() for x in kmers_batch[i]
                    axs.plot(rc_signal[i][:duration_batch[i]],color = 'blue',label = 'reconstructed hmm',alpha = 0.7)
                    axs.plot(rc_signal_kmer[i][:duration_batch[i]],color = 'yellow',label = 'reconstructed basecaller',alpha = 0.7)
                    # axs.plot(rc_signal_orig[:duration_batch[i]],color = "red",label = "original label")
                    plt.legend()
                    plt.xlabel("Time")
                    plt.ylabel("Signal")
                    plt.figure()
                    dwell_length = np.asarray([len(list(x)) for _,x in itertools.groupby(path[i][:duration_batch[i]])])
                    sns.histplot(dwell_length[dwell_length<10])
        return collection
            
    def proportion_test(self,
                        n_batch = 1):
        if self.loader is None:
            raise ValueError("No dataset has been loaded, call Tester.load_dataset first.")
        collection = {"m_prop":[],
                      "m_count":[],
                      "a_count":[],
                      "m_prop_aln":[],
                      "a_count_aln":[],
                      "m_count_aln":[]}
        for i_batch, batch in tqdm(enumerate(self.loader)):
            if i_batch >= n_batch:
                break
            duration_batch = batch['duration']
            transition_batch = batch['labels']
            kmers_batch = batch['kmers']
            renorm_batch,_,_ = self.renormalization(batch)
            with torch.no_grad():
                path,logit = self.hmm.viterbi_decode(renorm_batch, duration_batch, transition_batch)
            for i in np.arange(self.config.batch_size):
                deco_seq = kmers2seq(path.cpu().numpy()[i][:duration_batch[i]],self.data_config['idx2kmer'])
                orig_seq = kmers2seq(kmers_batch.cpu().numpy()[i][:duration_batch[i]],self.data_config['idx2kmer'])
                aligner = Methylation_DP_Aligner(base_alternation = {'M':'A'})
                deco_aln,orig_aln = aligner.align(deco_seq,orig_seq)
                final_seq = aligner.merge(deco_aln,orig_aln)
                collection["m_count"].append(deco_seq.count("M"))
                collection["a_count"].append(deco_seq.count("A"))
                collection["m_prop"].append(deco_seq.count('M')/(deco_seq.count('A')+deco_seq.count('M')))
                collection["m_count_aln"].append(final_seq.count('M'))
                collection["a_count_aln"].append(final_seq.count('A'))
                collection["m_prop_aln"].append(final_seq.count('M')/(final_seq.count('M')+final_seq.count('A')))
        return collection
    
    def auc_test(self,
                 n_batch = 1,
                 thresholds = []):
        if self.loader is None:
            raise ValueError("No dataset has been loaded, call Tester.load_dataset first.")
        collection = {"TP":[],"FP":[]}
        for i_batch, batch in tqdm(enumerate(self.loader)):
            if i_batch >= n_batch:
                break
            signal_batch = batch['signal']
            duration_batch = batch['duration']
            transition_batch = batch['labels']
            kmers_batch = batch['kmers']
            renorm_batch,rc_signal,rc_signal_kmer = self.renormalization(batch)
            rocs = {x:[] for x in thresholds}
            with torch.no_grad():
                gamma = self.hmm.expectation(renorm_batch.float(), duration_batch, transition_batch)
            for t in tqdm(thresholds):
                rocs[t]+= posterior_decode(gamma,self.data_config['idx2kmer'],self.data_config['kmer2idx_dict'],t)
        return rocs
    
if __name__ == "__main__":
    import seaborn as sns
    torch.manual_seed(1992)
    home_f = os.path.expanduser('~') 
    class ModelArguments:
        batch_size = 10
        device = "cuda" if torch.cuda.is_available() else "cpu"
        m_prior = 0.3 #The prior factor of a base being normalized, 1 means no bias.
        neighbour = 4
        ckpt_f = None
        ## Renormalization arguments
        dwell_norm = True
        no_scale = True
        n_renorm = 3
        fix_alignmetn = False
        
        ## arguments for old function
        visual = False
        n_cases = 1
        
    model_config = ModelArguments
    
    ### Models
    # ckpts = ['/home/heavens/bridge_scratch/NRHMM_models/xron_rhmm_models_new/ckpt-36234',
    #          '/home/heavens/bridge_scratch/NRHMM_models/xron_rhmm_models_new/ckpt-31613',
    #          '/home/heavens/bridge_scratch/NRHMM_models/xron_rhmm_models_new/ckpt-12855']
    ckpts = ['/home/heavens/bridge_scratch/NRHMM_models/rhmm_mm_norm/ckpt-8609']
    compare = {"m6A":home_f + "/bridge_scratch/ime4_Yearst/IVT/m6A/rep2/kmers_guppy_4000_noise",
                "control":home_f + "/bridge_scratch/NA12878_RNA_IVT/xron_output/kmers_xron_4000_noise/"}
    
    # compare = {"m6A":home_f + "/bridge_scratch/m6A_Nanopore_RNA002/data/m6A_25_pct/20210430_1751_X1_FAQ16555_e98b69f8/kmers_guppy_4000_noise"}
    # compare = [home_f + "/bridge_scratch/ime4_Yearst/IVT/m6A/rep1/kmers_guppy_4000_dwell",
    #             "/home/heavens/bridge_scratch/NA12878_RNA_IVT/xron_partial/extracted_kmers"]
    # compare = [home_f + "/bridge_scratch/NA12878_RNA_IVT/guppy_train/kmers_guppy_4000_dwell"]
    # compare = ["/home/heavens/bridge_scratch/NA12878_RNA_IVT/xron_partial/extracted_kmers"]
    # compare = ["/home/heavens/bridge_scratch/NA12878_RNA_IVT/xron_partial/extracted_kmers"]
    # compare = [home_f + "/bridge_scratch/ime4_Yearst/IVT/m6A/rep1/kmers_guppy_4000_dwell"]
    
    testers = []
    for ckpt in ckpts:
        model_config.ckpt_f = ckpt
        testers.append(Tester(model_config))
    
    # testers[0].add_dataset(compare["m6A"],name="P25")
    print("Adding datasets.")
    testers[0].add_dataset("/home/heavens/bridge_scratch/m6A_Nanopore_RNA002/data/m6A_25_pct/20210430_1751_X1_FAQ16555_e98b69f8/kmers_guppy_4000_noise",name = "P25")
    testers[0].add_dataset(home_f + "/bridge_scratch/NA12878_RNA_IVT/xron_output/kmers_xron_4000_noise/",name = "control")
    testers[0].add_dataset(home_f + "/bridge_scratch/ime4_Yearst/IVT/m6A/rep2/kmers_guppy_4000_noise",name = "m6Arep1")
    
    print("Datasets added successfully, loading the control dataset.")
    testers[0].load_dataset("control")
    
    print("Renormalization test.")
    summary = testers[0].renormalization_test(n_batch = 1,plot = True)
    
    print("Calculate ROC on control dataset.")
    roc_control = testers[0].auc_test(n_batch = 2, thresholds = np.arange(-100,100,0.5))
    
    print("Loading m6A dataset.")
    testers[0].load_dataset("m6Arep1")
    
    print("Calculate ROC on m6A dataset.")
    roc_meth = testers[0].auc_test(n_batch = 2, thresholds = np.arange(-100,100,0.5))
    
    # summary = testers[0].proportion_test(n_batch = 10)
    testers[0].load_dataset("m6Arep1")
    # summary_m6A = testers[0].proportion_test(n_batch = 10)
    TP = [np.mean(x) for x in roc_meth.values()]
    FP = [np.mean(x) for x in roc_control.values()]
    fig,axs = plt.subplots(figsize = (5,5))
    auc_plot(TP,FP,axs)
    fig,axs = plt.subplots(figsize = (5,5))
    TP_n,FP_n = multiple_testing(TP, FP, 10)
    auc_plot(TP_n,FP_n,axs)
    
